import re
import base64

from roscraco.router.base import RouterBase

from roscraco.helper import validator, converter
from roscraco.exception import RouterParseError, RouterIdentityError

from roscraco.response import RouterInfo, TrafficStats, DMZSettings, \
     ConnectedClientsListItem, ConnectedClientsList, WirelessSettings, \
     DHCPReservationList, DHCPReservationListItem, DHCPServerSettings


class Canyon_CNPWF514N1(RouterBase):

    def _perform_http_request(self, *args, **kwargs):
        auth = base64.b64encode('%s:%s' % (self.username, self.password))
        kwargs['headers'] = [('Authorization', 'Basic %s' % auth)]
        return RouterBase._perform_http_request(self, *args, **kwargs)

    def confirm_identity(self):
        _, headers, contents = self._perform_http_request('%sindex.asp' %
                                                          self.url_base)

        header_server = headers.getheader('Server')
        if header_server != 'GoAhead-Webs':
            raise RouterIdentityError('Invalid Server header: %s' % header_server)

        string_to_find = '/file/multilanguage.var'
        if string_to_find not in contents:
            raise RouterIdentityError('Cannot find string in contents: %s' % string_to_find)

    def get_router_info(self):
        return _parse_router_info(self._make_http_request_read('stainfo.asp'))

    def get_uptime(self):
        return _parse_uptime(self._make_http_request_read('stainfo.asp'))

    def get_pppoe_online_time(self):
        return None

    def get_traffic_stats(self):
        return _parse_traffic_stats(self._make_http_request_read('stats.asp'))

    def get_mac_address(self):
        return _parse_mac_address(self._make_http_request_read('stanet.asp'))

    def get_dns_servers(self):
        return _parse_dns_servers(self._make_http_request_read('stanet.asp'))

    def get_connected_clients_list(self):
        # What's interesting about this router is that the DHCP list
        # shows LAN clients only (and not Wi-Fi clients)
        # There's a separate list for Wi-Fi clients, but it shows only IPs,
        # so it's worthless..
        # The DHCP server (and address reservations) work on the LAN only!
        html = self._make_http_request_read('stadhcptbl.asp')
        return _parse_connected_clients_list(html)

    def get_dmz_settings(self):
        # DMZ not supported at all
        settings = DMZSettings()
        settings.set_supported_status(False)
        settings.set_enabled_status(False)
        settings.set_ip('0.0.0.0')
        return settings

    def push_dmz_settings(self, settings):
        pass

    def get_dhcp_settings(self):
        return _parse_dhcp_settings(self._make_http_request_read('lan.asp'))

    def get_addr_reservation_list(self):
        return _parse_addr_reservations(self._make_http_request_read('lan.asp'))

    def push_addr_reservation_list(self, lst_new):
        lst_new.ensure_valid()

        lst_old = self.get_addr_reservation_list()

        if lst_old == lst_new:
            return True

        if len(lst_old) != 0:
            # delete all old entries at once, before we push the new entries
            data = _generate_addr_reservation_delete_data(lst_old)
            self._make_http_request_write('goform/formSDHCP', data)

        for item in lst_new:
            data = _generate_addr_reservation_add_data(item)
            self._make_http_request_write('goform/formSDHCP', data)

        return self.get_addr_reservation_list() == lst_new

    def get_wireless_settings(self):
        html = []
        html.append(self._make_http_request_read('wlmain.asp'))
        html.append(self._make_http_request_read('wlbasic.asp'))
        html.append(self._make_http_request_read('wladvance.asp'))
        html.append(self._make_http_request_read('wlencrypt.asp'))
        return _parse_wireless_settings(*html)

    def push_wireless_settings(self, settings):
        data_basic = _generate_wireless_data_basic(settings)
        self._make_http_request_write('goform/formWlanSetup', data_basic)

        data_advanced = _generate_wireless_data_advanced(settings)
        self._make_http_request_write('goform/formAdvanceSetup', data_advanced)

        data_security = _generate_wireless_data_security(settings)
        self._make_http_request_write('goform/formWlEncrypt', data_security)

        # We MUST update this last.. because it potentially DISABLES wireless
        # If we disable it, and push the above settings after that
        # the router will act smart and re-enable wireless again
        data_main = _generate_wireless_data_main(settings)
        self._make_http_request_write('goform/formWlEnable', data_main)

        settings_now = self.get_wireless_settings()

        if not settings.security_type_is_wep:
            return settings_now == settings

        # We're changing to WEP security, and we can't extract WEP passwords
        # so we need some custom validation
        return settings_now.eq(settings, skip_attrs=('password', )) and settings_now.password is None

    @property
    def supports_reboot(self):
        return True

    def reboot(self):
        data = {'reset_flag': 0, 'submit-url': '/tlreset.asp'}
        self._make_http_request_write('goform/formReboot', data)


def _parse_router_info(html):
    info = RouterInfo()

    match_object = re.compile("dw\(HardwareVersion\)(?:.+?)table2(?:.+?)>&nbsp;(.+?)</td>", re.DOTALL).search(html)
    if match_object is None:
        raise RouterParseError('Cannot determine hardware version')
    info.set_hardware_version(match_object.group(1))

    match_object = re.compile("dw\(RuntimeCodeVersion\)(?:.+?)table2(?:.+?)>&nbsp;(.+?)</td>", re.DOTALL).search(html)
    if match_object is None:
        raise RouterParseError('Cannot determine hardware version')
    info.set_firmware_version(match_object.group(1))

    return info


def _parse_uptime(html):
    regex = re.compile("dw\(UpTime\)(?:.+?)table2(?:.+?)&nbsp;([0-9]+)day:([0-9]+)h:([0-9]+)m:([0-9]+)s</td>", re.DOTALL)
    match_object = regex.search(html)
    if match_object is None:
        raise RouterParseError('Cannot _parse uptime')

    days, hours, minutes, seconds = map(int, match_object.groups())

    return days * 86400 + hours * 3600 + minutes * 60 + seconds


def _parse_traffic_stats(html):
    regex = "showText\(EthernetWAN\)(?:.+?)table2(?:.+?)&nbsp;([0-9]+)(?:.+?)table2(?:.+?)&nbsp;([0-9]+)"
    match_object = re.compile(regex, re.DOTALL).search(html)
    if match_object is None:
        raise RouterParseError('Cannot _parse traffic stats')

    packets_sent, packets_recv = map(int, match_object.groups())
    # sadly, we've got only packets and no size information..
    return TrafficStats(0, 0, packets_recv, packets_sent)


def _parse_mac_address(html):
    match_object = re.compile("dw\(MACAddress\)(?:.+?)table2(?:.+?)>&nbsp;(.+?)\s", re.DOTALL).search(html)
    if match_object is None:
        raise RouterParseError('Cannot determine mac address')

    mac = match_object.group(1).strip(" ")
    if not validator.is_valid_mac_address(mac):
        raise RouterParseError('Found an invalid MAC address: %s' % mac)

    return converter.normalize_mac(mac)


def _parse_dns_servers(html):
    regex_dns_servers = re.compile('name="dns[1,2]" value="(.+?)\s')
    servers = re.findall(regex_dns_servers, html)
    return [ip.strip(" ") for ip in servers if validator.is_valid_ip_address(ip)]


def _parse_connected_clients_list(html):
    regex_dhcp_list = "<tr class=table2 align=center><td><font size=2>(.+?)</td><td><font size=2>(.+?)</td>"
    lst = ConnectedClientsList()

    for id, (ip, mac) in enumerate(re.findall(regex_dhcp_list, html), start=1):
        if ip == "None":
            # this entry is added when there are no connected clients
            break

        if not validator.is_valid_ip_address(ip):
            raise RouterParseError('Invalid IP address: %s' % ip)

        if not validator.is_valid_mac_address(mac):
            raise RouterParseError('Invalid MAC address: %s' % mac)

        item = ConnectedClientsListItem()
        item.set_client_name('Client %d' % id)
        item.set_mac(converter.normalize_mac(mac))
        item.set_ip(ip)
        item.set_lease_time(0)

        lst.append(item)

    return lst


def _parse_dhcp_settings(html):
    regex = re.compile('name="dhcpRange(?:Start|End)" size="15" maxlength="15" value="(.+?)"')
    try:
        ip_start, ip_end = regex.findall(html)
    except ValueError:
        raise RouterParseError('Cannot find DHCP start/end range')
    else:
        settings = DHCPServerSettings()
        settings.set_ip_start(ip_start)
        settings.set_ip_end(ip_end)
        settings.set_enabled_status('<option selected value="2"><script>dw(Enable)</script></option>' in html)
        settings.ensure_valid()

        return settings


def _parse_addr_reservations(html):
    regex = re.compile('name="smac[0-9]+" value="(.+?)"(?:.+?)name="sip[0-9]+" value="(.+?)">', re.DOTALL)

    reservation_list = DHCPReservationList()
    reservation_list.set_reboot_requirement_status(False)

    for mac, ip in regex.findall(html):
        item = DHCPReservationListItem()
        item.set_mac(converter.normalize_mac(mac))
        item.set_ip(ip)
        item.set_enabled_status(True)
        reservation_list.append(item)
    return reservation_list


def _parse_wireless_settings(html_main, html_basic, html_advanced, html_security):
    obj = WirelessSettings()
    obj.add_security_support(WirelessSettings.SECURITY_TYPE_WEP64)
    obj.add_security_support(WirelessSettings.SECURITY_TYPE_WEP128)
    obj.add_security_support(WirelessSettings.SECURITY_TYPE_WPA)
    obj.add_security_support(WirelessSettings.SECURITY_TYPE_WPA2)

    obj.set_auto_channel_support(False)

    # the naming of the radio button is weird!
    is_enabled = '<input type="radio" name="wlanDisabled" value="yes" checked>' in html_main
    obj.set_enabled_status(is_enabled)


    match_object = re.compile('document.wlanSetup.ssid.value="(.+?)";').search(html_basic)
    if match_object is None:
        raise RouterParseError('Cannot determine SSID')
    obj.set_ssid(match_object.group(1))

    match_object = re.compile('var defaultChan = ([0-9]+)[\s\n]').search(html_basic)
    if match_object is None:
        raise RouterParseError('Cannot determine channel')
    obj.set_channel(int(match_object.group(1)))

    is_broadcasting_ssid = '<input type="radio" name="hiddenSSID" value="no" checked>' in html_advanced
    obj.set_ssid_broadcast_status(is_broadcasting_ssid)

    match_object = re.compile('methodVal = ([0-3]);').search(html_security)
    if match_object is None:
        raise RouterParseError('Cannot determine security type')
    security_type = int(match_object.group(1))
    if security_type == 0:
        obj.set_security_type(obj.__class__.SECURITY_TYPE_NONE)
    elif security_type == 1: # WEP of some sort
        match_object = re.compile('var wepTbl =\s+new Array\("([0-9]+)"\);').search(html_security)
        if match_object is None:
            raise RouterParseError('Cannot determine WEP bit length')
        if int(match_object.group(1)) in (0, 1):
            obj.set_security_type(obj.__class__.SECURITY_TYPE_WEP64)
        else:
            obj.set_security_type(obj.__class__.SECURITY_TYPE_WEP128)

        # WEP passwords cannot be extracted! It shows "***" only
    elif security_type == 2: # WPA of some sort
        match_object = re.compile('var wpaCipherTbl =(?:\s+)?new Array\("([0-9]+)"\);').search(html_security)
        if match_object is None:
            raise RouterParseError('Cannot determine WPA type')
        if int(match_object.group(1)) in (0, 1, 3): # 0, 1 = WPA; 3 = mixed
            obj.set_security_type(obj.__class__.SECURITY_TYPE_WPA)
        else: # all other values are WPA2 only
            obj.set_security_type(obj.__class__.SECURITY_TYPE_WPA2)

        match_object = re.compile('var pskValueTbl = new Array\("(.+?)"\);').search(html_security)
        if match_object is None:
            raise RouterParseError('Cannot determine WPA password')
        obj.set_password(match_object.group(1))
    else: # 3 = WPA Radius
        raise NotImplementedError('Security type not supported')

    return obj


def _generate_addr_reservation_delete_data(lst_old):
    if len(lst_old) == 0:
        raise Exception('Nothing to delete!')

    data = {}
    data['deleteAllDhcpMac'] = 'Delete All'
    data['submit-url'] = '/lan.asp'
    data['smacnum'] = len(lst_old)

    for i, item in enumerate(lst_old, start=1):
        data['sip%d' %i] = item.ip
        data['smac%d' %i] = item.mac

    return data


def _generate_addr_reservation_add_data(item):
    data = {}
    data['SDHCPEnabled'] = 'ON'
    data['addSDHCPMac'] = 'Add'
    data['submit-url'] = '/lan.asp'
    data['tiny_idx'] = 0
    data['ip'] = item.ip
    data['mac'] = item.mac
    return data


def _generate_wireless_data_main(settings):
    settings.ensure_valid()

    data = {'wlan-url': '/wlmain.asp'}

    # yes.. this seems backwards! bad field/value naming on Canyon's part!
    data['wlanDisabled'] = 'yes' if settings.is_enabled else 'no'

    return data


def _generate_wireless_data_basic(settings):
    settings.ensure_valid()

    data = {'B1': 'APPLY', 'wlan-url': '/wlbasic.asp'}
    data['apMode'] = '0' # 0 = Access Point (normal mode..)

    data['band'] = 4 # 0 = B; 1 = N; 2 = B+G; 3 = G; 4 = B+G+N
    data['chan'] = settings.channel
    data['ssid'] = settings.ssid

    # Repeater-Mode related settings.. we don't support this
    data['repeaterSSID'] = ''
    data['autoMacClone'] = 'no'
    data['wisp'] = 0
    data['macAddrValue'] = ''

    # Universal Wireless Repeater related stuff..
    data['wlLinkMac1'] = '000000000000'
    data['wlLinkMac2'] = '000000000000'
    data['wlLinkMac3'] = '000000000000'
    data['wlLinkMac4'] = '000000000000'
    data['wlanMac'] = '000000000000'

    # This should be 0 if the SSID remains the same
    # and 1 if it has changed..
    # Unfortunately, we don't know what the previous value is,
    # so let's supply 1.. it seems to work fine
    data['wpsStatus'] = 1

    return data


def _generate_wireless_data_advanced(settings):
    settings.ensure_valid()

    data = {}

    # We're basically using "sane" settings everywhere
    # and are only changing the "Broadcast SSID" field
    # Things could be improved if the original settings were preserved
    # we can do that by setting "internal_params" on the WirelessSettings object
    # and using their original values here
    # .. maybe that's something @todo

    data['B1'] = 'APPLY'
    data['NtxRate'] = 0
    data['beaconInterval'] = 100
    data['disProtection'] = 'no'
    data['dtimPeriod'] = 3
    data['fragThreshold'] = 2346
    data['getRate'] = 'auto'
    data['getTxPower'] = 0
    data['hiddenSSID'] = 'no' if settings.is_broadcasting_ssid else 'yes'
    data['iapp'] = 'yes'
    data['ipnPrmb'] = 0
    data['preamble'] = 'long'
    data['rtsThreshold'] = 2347
    data['submit-url'] = '/wladvance.asp'
    data['txRate'] = 0
    data['wlanBurst'] = 'no'
    data['wlanCts'] = 'none'
    data['wlanNChanWidth'] = 0
    data['wlanRateMode'] = 'mixed'
    data['wlanTxPower'] = 0
    data['wlanWmm'] = 'no'

    return data


def _generate_wireless_data_security(settings):
    settings.ensure_valid()

    data = {}

    # Set default WEP keys.. to be changed later if in WEP mode
    data['key1'], data['key2'], data['key3'], data['key4'] = ('*' * 10,) * 4

    # Specify that the first WEP key is to be used
    data['defaultTxKeyId'] = 1

    if settings.security_type_is_wep:
        data['format'] = 2 if settings.is_wep_password_in_hex else 1
        data['key1'] = settings.password

        # bit length (64bit vs 128bit)
        data['length'] = 1 if settings.security_type == settings.__class__.SECURITY_TYPE_WEP64 else 2

        data['method'] = 1 # security_type = WEP
    else:
        data['format'] = 2
        data['length'] = 1

        if settings.security_type_is_wpa:
            data['method'] = 2 # for both WPA and WPA2
        else: # security is disabled or unsupported
            data['method'] = 0


    data['pskFormat'] = 0 # 0 = ASCII, 1 = HEX
    data['pskValue'] = settings.password

    # Unsupported WPA Radius auth settings
    data['radiusIP'], data['radiusPass'], data['radiusPort'] = '', '', 1812
    data['sel1xMode'] = 'ON'

    data['selected_ssid'] = 0 # has something to do with SSID presets and switching

    data['submit-url'] = '/wlencrypt.asp'

    # This is weird.. I've even seen it with WPA2
    data['wepEnabled'] = 'ON'

    # 1 = WPA; 2 = WPA2; 3 = WPA2 Mixed
    data['wpaCipher'] = 1 if settings.security_type == settings.__class__.SECURITY_TYPE_WPA else 2

    return data




class TrafficStats(object):
    """Represents traffic statistics information for a given interface."""

    def __init__(self, bytes_recv, bytes_sent, packets_recv, packets_sent):
        self._bytes_recv = bytes_recv
        self._bytes_sent = bytes_sent
        self._packets_recv = packets_recv
        self._packets_sent = packets_sent

    @property
    def bytes_recv(self):
        return self._bytes_recv

    @property
    def bytes_sent(self):
        return self._bytes_sent

    @property
    def packets_recv(self):
        return self._packets_recv

    @property
    def packets_sent(self):
        return self._packets_sent




from roscraco.helper import validator
from roscraco.exception import RouterSettingsError


class WirelessSettings(object):
    """Represents all available Wireless settings for a router."""

    SECURITY_TYPE_NONE = 'none'
    SECURITY_TYPE_WEP64 = 'wep64'
    SECURITY_TYPE_WEP128 = 'wep128'
    SECURITY_TYPE_WPA = 'wpa'
    SECURITY_TYPE_WPA2 = 'wpa2'

    #: List of properties to export using export()
    PROPERTIES = (
        'security_type', 'ssid', 'is_enabled', 'is_broadcasting_ssid',
        'channel', 'password'
    )

    def __init__(self):
        self._supports_wireless = True
        self._ssid = None
        self._enabled_status = True
        self._ssid_broadcast_status = True
        self._channel = None
        self._password = None
        self._internal_params = {}
        self._supported_security_types = set([self.__class__.SECURITY_TYPE_NONE])
        self._security_type = None
        self._supports_ascii_wep_passwords = True
        self._supports_auto_channel = True
        self._changes_require_reboot = True

    def set_auto_channel_support(self, value):
        self._supports_auto_channel = bool(value)

    @property
    def supports_auto_channel(self):
        """Tells whether auto channel is supported.

        Channel 0 is considered the auto channel, because that's
        how most routers represent the ``Auto`` value.

        Some devices, however, do not support Auto channel at all.
        """
        return self._supports_auto_channel

    def add_security_support(self, security_type):
        """Adds a new security type to the list of supported
        security types.
        """
        self._supported_security_types.add(security_type)

    @property
    def supported_security_types(self):
        return self._supported_security_types

    def set_security_type(self, security_type):
        self._security_type = security_type

    @property
    def security_type_is_wep(self):
        """Tells whether the current security type is WEP.

        Returns true for both WEP64 and WEP128.
        """
        return self._security_type in (self.__class__.SECURITY_TYPE_WEP64, self.__class__.SECURITY_TYPE_WEP128)

    @property
    def security_type_is_wpa(self):
        """Tells whether the current security type is WPA.

        Returns true for both WPA and WPA2.
        """
        return self._security_type in (self.__class__.SECURITY_TYPE_WPA, self.__class__.SECURITY_TYPE_WPA2)

    @property
    def security_type(self):
        return self._security_type

    def set_reboot_requirement_status(self, value):
        self._changes_require_reboot = bool(value)

    @property
    def changes_require_reboot(self):
        """Tells whether the router needs rebooting
        for changes to take effect.
        """
        return self._changes_require_reboot

    def set_support_status(self, value):
        self._supports_wireless = bool(value)

    @property
    def is_supported(self):
        """Tells whether the router supports wireless (most of them do)."""
        return self._supports_wireless

    def set_ssid(self, value):
        self._ssid = value

    @property
    def ssid(self):
        """The current SSID (wireless network name)."""
        return self._ssid

    def set_enabled_status(self, value):
        self._enabled_status = bool(value)

    @property
    def is_enabled(self):
        return self._enabled_status

    def set_ssid_broadcast_status(self, value):
        self._ssid_broadcast_status = bool(value)

    @property
    def is_broadcasting_ssid(self):
        """Tells whether the SSID status is being broadcasted publicly.

        If it is, than the network is publicly visible by anyone.
        """
        return self._ssid_broadcast_status

    def set_channel(self, value):
        self._channel = int(value)

    @property
    def channel(self):
        """The transmission channel for wireless communications."""
        return self._channel

    def set_password(self, value):
        self._password = value

    @property
    def password(self):
        """The current password for the given security type.

        The password is sometimes None for some routers, to indicate
        that the password cannot be determined.
        Some routers hide the current password from their web-interface,
        so we can't detect it (but that doesn't mean that we can't change it
        with a new one).
        """
        return self._password

    @property
    def is_wep_password_in_hex(self):
        """Tells whether the current WEP password is in HEX or in ASCII.

        Detecting this allows us to set the ASCII/HEX
        field in the management interface automatically.
        """
        if not self.security_type_is_wep:
            raise RouterSettingsError('Not using WEP, but trying to inspect password!')
        bit_length = 128 if self.security_type == self.__class__.SECURITY_TYPE_WEP128 else 64
        return validator.is_wep_password_in_hex(self.password, bit_length)

    def set_ascii_wep_password_support_status(self, value):
        self._supports_ascii_wep_passwords = bool(value)

    @property
    def supports_ascii_wep_passwords(self):
        """Tells whether the current router supports ASCII passwords
        for WEP security.

        Some devices only support HEX passwords.
        """
        return self._supports_ascii_wep_passwords

    def set_internal_param(self, key, value):
        self._internal_params[key] = value

    def get_internal_param(self, key):
        return self._internal_params[key] if key in self._internal_params else None

    def validate(self):
        errors = {}

        if not validator.is_valid_ssid(self.ssid):
            errors['ssid'] = 'Invalid SSID: %s' % self.ssid

        # most routers use channel 0 as the 'Auto' channel
        channel_min = 0 if self.supports_auto_channel else 1
        if not (channel_min <= self.channel <= 13):
            errors['channel'] = 'Invalid channel %d' % self.channel

        if self.security_type not in self._supported_security_types:
            errors['security_type'] = 'Invalid security type: %s' % self.security_type
        else:
            result = self.__validate_password()
            if result is not None:
                errors['password'] = result

        return errors

    def ensure_valid(self):
        errors = self.validate()
        if len(errors) != 0:
            raise RouterSettingsError(str(errors))

    def __validate_password(self):
        if self.security_type in (self.__class__.SECURITY_TYPE_WPA, self.__class__.SECURITY_TYPE_WPA2):
            if not validator.is_valid_wpa_psk_password(self.password):
                return 'Invalid WPA PSK password: %s' % self.password

        if self.security_type in (self.__class__.SECURITY_TYPE_WEP64, self.__class__.SECURITY_TYPE_WEP128):
            bit_length = 128 if self.security_type == self.__class__.SECURITY_TYPE_WEP128 else 64
            if not validator.is_valid_wep_password(self.password, bit_length):
                return 'Invalid WEP password for bit length %d: %s' % (bit_length, self.password)

            # Some devices only support HEX values for the WEP password field
            if not self.supports_ascii_wep_passwords and not self.is_wep_password_in_hex:
                return 'ASCII WEP passwords are not supported!'

        return None

    def eq(self, other, skip_attrs=()):
        # WEP passwords that use HEX are not case-sensitive, so we want
        # to validate them separately
        if self.security_type_is_wep and other.security_type_is_wep and \
           self.is_wep_password_in_hex and other.is_wep_password_in_hex:
            skip_attrs = skip_attrs + ('password',)
            try:
                if self.password.lower() != other.password.lower():
                    return False
            except AttributeError:
                return False

        # Don't try to compare passwords when there's no security type
        if self.security_type == self.__class__.SECURITY_TYPE_NONE and \
           other.security_type == self.__class__.SECURITY_TYPE_NONE:
               skip_attrs = skip_attrs + ('password',)

        for attr in self.__class__.PROPERTIES:
            if attr in skip_attrs:
                continue
            if getattr(self, attr, None) != getattr(other, attr, None):
                #print('[%s] %s != %s' % (
                #               attr,
                #               getattr(self, attr, None),
                #               getattr(other, attr, None)
                #))
                return False

        return True

    def __eq__(self, other):
        return self.eq(other)

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return id(self)

    def export(self):
        """Exports the most important settings attributes,
        omitting any internal attributes.
        """
        export = {}
        for attr in self.__class__.PROPERTIES:
            export[attr] = getattr(self, attr, None)
        return export




import re
import base64

from roscraco.router.base import RouterBase

from roscraco.helper import validator, converter
from roscraco.exception import RouterFetchError, RouterParseError, \
     RouterIdentityError

from roscraco.response import RouterInfo, TrafficStats, DMZSettings, \
     ConnectedClientsListItem, ConnectedClientsList, WirelessSettings, \
     DHCPReservationList, DHCPReservationListItem, DHCPServerSettings


class Zyxel_P330W(RouterBase):

    def _perform_http_request(self, *args, **kwargs):
        auth = base64.b64encode('%s:%s' % (self.username, self.password))
        kwargs['headers'] = [('Authorization', 'Basic %s' % auth)]
        return RouterBase._perform_http_request(self, *args, **kwargs)

    def _wait_for_settings_reload(self, wait_time=32):
        """Blocks until the router reloads its settings
        and becomes ready to respond to requests again.
        """
        from time import sleep
        sleep(wait_time)

    def confirm_identity(self):
        try:
            html = self._make_http_request_read('status.asp')
            if '<title>Wireless Router Status</title>' not in html:
                raise RouterIdentityError(
                    'Cannot confirm that this is a router!'
                )
            # would raise RouterParseError if it fails
            self.get_router_info()
        except (RouterParseError, RouterFetchError):
            raise RouterIdentityError('Cannot confirm identity!')

    def get_router_info(self):
        return _parse_router_info(self._make_http_request_read('status.asp'))

    def get_uptime(self):
        return None

    def get_pppoe_online_time(self):
        return None

    def get_traffic_stats(self):
        html = self._make_http_request_read('stats.asp')
        return _parse_traffic_stats(html)

    def get_mac_address(self):
        html = self._make_http_request_read('status.asp')
        return _parse_mac_address(html)

    def get_dns_servers(self):
        # We cannot determine DNS servers that are automatically
        # assigned (for PPPoE connections), so this can only be made
        # to work for manually added DNS server entries.
        # These are rare cases, so we won't bother for now.
        return []

    def get_connected_clients_list(self):
        html = self._make_http_request_read('dhcptbl.asp')
        return _parse_connected_clients_list(html)

    def get_dmz_settings(self):
        html = self._make_http_request_read('dmz.asp')
        return _parse_dmz_settings(html)

    def push_dmz_settings(self, settings):
        settings_old = self.get_dmz_settings()
        if settings == settings_old:
            return True

        data = _generate_dmz_data(settings)
        # The router reloads its settings as we make the request,
        # so it's normal for it to take about 10 seconds
        self._make_http_request_write('goform/formDMZ', data, timeout=15.0)

        new = self.get_dmz_settings()
        if settings.is_enabled:
            # We can push a full update when it's enabled
            return new == settings
        else:
            # We can only switch the enabled status, but
            # can't change the IP when disabling..
            return settings.is_enabled == new.is_enabled

    def get_dhcp_settings(self):
        html = self._make_http_request_read('tcpiplan.asp')
        return _parse_dhcp_settings(html)

    def get_addr_reservation_list(self):
        html = self._make_http_request_read('tcpiplan.asp')
        return _parse_addr_reservation_list(html)

    def push_addr_reservation_list(self, lst_new):
        lst_new.ensure_valid()

        lst_old = self.get_addr_reservation_list()
        if lst_old == lst_new:
            return True

        if len(lst_old) != 0:
            delete = {}
            delete['delallLease'] = 'Delete All'
            delete['submit-url'] = '/tcpiplan.asp'
            delete['submit-value'] = '0.031'
            self._make_http_request_write('goform/formStaticLease', delete)
            self._wait_for_settings_reload()

        for item in lst_new:
            data = _generate_addr_reservation_item_data(item)
            self._make_http_request_write('goform/formStaticLease', data)
            self._wait_for_settings_reload()

        return self.get_addr_reservation_list() == lst_new

    def get_wireless_settings(self):
        html = [self._make_http_request_read(path) for path in (
            'wlbasic.asp', 'wladvanced.asp', 'wlwpa.asp', 'wlwep.asp'
        )]
        return _parse_wireless_settings(*html)

    def push_wireless_settings(self, settings):
        # Each time a settings update is pushed (to whatever page)
        # the router reloads its settings (which takes a lot of time).
        # It was possible to do multiple pushes one after another and
        # get by with only one settings reload, but one of the test routers
        # crashed when that was done. So we'll play it safe and wait each time.
        # To get around the slow updates, we only push updates to a page
        # that has changes.
        settings_old = self.get_wireless_settings()

        if settings.channel != settings_old.channel or \
           settings.ssid != settings_old.ssid:
            basic = _generate_wireless_data_basic(settings)
            self._make_http_request_write('goform/formWlanSetup', basic)
            self._wait_for_settings_reload()

        if settings.is_broadcasting_ssid != settings_old.is_broadcasting_ssid:
            advanced = _generate_wireless_data_advanced(settings)
            self._make_http_request_write('goform/formAdvanceSetup', advanced)
            self._wait_for_settings_reload()

        if settings.security_type_is_wep and \
           settings.password != settings_old.password:
            wep = _generate_wireless_data_wep(settings)
            self._make_http_request_write('goform/formWep', wep)
            self._wait_for_settings_reload()

        if settings.security_type != settings_old.security_type or \
           (settings.security_type_is_wpa and \
            settings.password != settings_old.password):
            security = _generate_wireless_data_security(settings)
            self._make_http_request_write('goform/formWlEncrypt', security)
            self._wait_for_settings_reload()

        return settings.eq(self.get_wireless_settings(), skip_attrs=('password',))

    @property
    def supports_reboot(self):
        return False

    def reboot(self):
        pass

def _parse_router_info(html):
    obj = RouterInfo()
    obj.set_hardware_version('P-330W EE')
    match_object = re.compile('>(P-330W_EE_(?:.+?))</font>').search(html)
    if match_object is None:
        raise RouterParseError('Cannot determine firmware version!')
    obj.set_firmware_version(match_object.group(1))
    return obj


def _parse_traffic_stats(html):
    regex = re.compile('Ethernet WAN(?:.+?)<td>([0-9]+)</td>'
                       '(?:.+?)<td>([0-9]+)</td>', re.DOTALL)
    match_object = regex.search(html)
    if match_object is None:
        raise RouterParseError('Cannot determine WAN stats!')
    packets_sent, packets_recv = map(int, match_object.groups())
    return TrafficStats(0, 0, packets_recv, packets_sent)


def _parse_mac_address(html):
    regex = re.compile('<!--Wan Information-->(?:.+?)Physical Address(?:.+?)'
                       '>((?:[a-f0-9]{2}:){5}[a-f0-9]{2})</font>', re.DOTALL)
    match_object = regex.search(html)
    if match_object is None:
        raise RouterParseError('Cannot determine MAC address!')
    return converter.normalize_mac(match_object.group(1))


def _parse_connected_clients_list(html):
    lst = ConnectedClientsList()

    regex = re.compile('<tr bgcolor=#(?:.+?)><td><font size=2>(.+?)</td>'
                       '<td><font size=2>(.+?)</td>'
                       '<td><font size=2>([0-9]+)</td></tr>')
    for i, (ip, mac, time) in enumerate(regex.findall(html)):
        item = ConnectedClientsListItem()
        item.set_client_name('Client %d' % (i + 1))
        item.set_ip(ip)
        item.set_mac(converter.normalize_mac(mac))
        item.set_lease_time(int(time))
        lst.append(item)
    return lst


def _parse_dmz_settings(html):
    settings = DMZSettings()
    settings.set_reboot_requirement_status(False)
    settings.set_supported_status(True)
    enabled = '<input type=checkbox name="enabled" value="ON" checked' in html
    settings.set_enabled_status(enabled)
    regex = re.compile('<input type=text name="ip" size=15'
                      ' maxlength=16 value="(.+?)"')
    match_object = regex.search(html)
    if match_object is None:
        raise RouterParseError('Cannot determine DMZ IP!')
    ip = match_object.group(1)
    if not validator.is_valid_ip_address(ip):
        if not enabled:
            ip = '192.168.1.1' # sane default value
        else:
            raise RouterParseError('Invalid IP address!')
    settings.set_ip(ip)
    return settings


def _parse_dhcp_settings(html):
    is_enabled = 'var choice= 2 ;' in html

    regex = re.compile('<input type=text name="dhcpRangeStart" size=16 '
                       'maxlength=15 value="(.+?)"'
                       '(?:.+?)'
                       '<input type=text name="dhcpRangeEnd" size=16 '
                       'maxlength=15 value="(.+?)"', re.DOTALL)
    match_object = regex.search(html)
    if match_object is None:
        raise RouterParseError('Cannot determine DMZ range!')
    for ip in match_object.groups():
        if not validator.is_valid_ip_address(ip):
            raise RouterParseError('Invalid DHCP IP: %s' % ip)

    ip_start, ip_end = match_object.groups()
    obj = DHCPServerSettings()
    obj.set_enabled_status(is_enabled)
    obj.set_ip_start(ip_start)
    obj.set_ip_end(ip_end)
    obj.ensure_valid()
    return obj


def _parse_addr_reservation_list(html):
    lst = DHCPReservationList()
    lst.set_reboot_requirement_status(False)
    regex = re.compile('<tr><td align=center width="50%" bgcolor="#C0C0C0">'
                       '<font size="2">(.+?)</td>(?:.+?)'
                       '<font size="2">(.+?)</td>(?:.+?)'
                       'name="select[0-9]+"', re.DOTALL)
    for mac, ip in regex.findall(html):
        item = DHCPReservationListItem()
        item.set_mac(mac)
        item.set_ip(ip)
        lst.append(item)
    return lst


def _parse_wireless_settings(html_basic, html_advanced, html_security, html_wep):
    settings = WirelessSettings()
    settings.add_security_support(WirelessSettings.SECURITY_TYPE_WEP64)
    settings.add_security_support(WirelessSettings.SECURITY_TYPE_WEP128)
    settings.add_security_support(WirelessSettings.SECURITY_TYPE_WPA)
    settings.add_security_support(WirelessSettings.SECURITY_TYPE_WPA2)
    settings.set_reboot_requirement_status(False)

    markup = '<input type=checkbox name="wlanDisabled0" value="ON" checked'
    settings.set_enabled_status(markup not in html_basic)

    regex = re.compile('defaultChan\[wlan_idx\]=([0-9]+);')
    match_object = regex.search(html_basic)
    if match_object is None:
        raise RouterParseError('Cannot determine channel.')
    settings.set_channel(int(match_object.group(1)))

    regex = re.compile('<input type=text name="ssid0" size=33 '
                       'maxlength=32 value="(.+?)"')
    match_object = regex.search(html_basic)
    if match_object is None:
        raise RouterParseError('Cannot determine SSID.')
    settings.set_ssid(match_object.group(1))

    markup = '<input type=radio name="hiddenSSID" value="no"checked>'
    settings.set_ssid_broadcast_status(markup in html_advanced)

    if '<option selected value=1>WEP' in html_security:
        passwords = re.compile('form.key10.value = "(.*?)";').findall(html_wep)
        # We expect 3 matches: 1) irrelevant 2) wep128 pass 3) wep64 pass
        if len(passwords) != 3:
            raise RouterParseError('Wrong number of passwords retrieved:'
                                   ' %d' % len(passwords))
        key_len_64 = '<input type=radio name="wepKeyLen0" value="wep64" checked>'
        if key_len_64 in html_security:
            settings.set_security_type(WirelessSettings.SECURITY_TYPE_WEP64)
            settings.set_password(passwords[2])
        else:
            settings.set_security_type(WirelessSettings.SECURITY_TYPE_WEP128)
            settings.set_password(passwords[1])
    elif '<option selected value=2>WPA' in html_security:
        settings.set_security_type(WirelessSettings.SECURITY_TYPE_WPA)
    elif '<option selected value=4>WPA2' in html_security:
        settings.set_security_type(WirelessSettings.SECURITY_TYPE_WPA2)
    else:
        settings.set_security_type(WirelessSettings.SECURITY_TYPE_NONE)

    return settings


def _generate_wireless_data_basic(settings):
    settings.ensure_valid()
    data = {}
    data['wlan-url'] = '/wlbasic.asp'
    data['submit-value'] = '0.031' # i don't know what it's for
    data['save'] = 'Save'
    data['basicrates0'] = 0 # i don't know what it's for
    data['operrates0'] = 0 # i don't know what it's for
    data['band0'] = 2 # 2 means B+G mode
    data['mode0'] = 0 # 0 means 'operate in AccessPoint mode'
    data['ssid0'] = settings.ssid
    data['chan0'] = settings.channel
    data['txRate'] = 0 # data rate = Auto
    data['wifiTestEnabled0'] = ''
    if not settings.is_enabled:
        data['wlanDisabled0'] = 'ON'
    return data


def _generate_wireless_data_advanced(settings):
    settings.ensure_valid()

    data = {}
    data['authType'] = 'both' # Authentication Type (could be open/shared/both)
    data['fragThreshold'] = 2346 # Fragment Threshold (int from 256 to 2346)
    data['rtsThreshold'] = 2347 # RTS Threshold (int from 0 to 2347)
    data['beaconInterval'] = 100 # Beacon interval in milliseconds
    data['preamble'] = 'long' # Preamble Type (could be long/short)
    data['hiddenSSID'] = 'no' if settings.is_broadcasting_ssid else 'yes'
    data['iapp'] = 'yes' # IAPP (could be yes/no)
    data['11g_protection'] = 'no' # 802.11g Protection (could be yes/no)
    data['save'] = 'Save'
    data['submit-url'] = '/wladvanced.asp'
    data['submit-value'] = '0.031'
    return data


def _generate_wireless_data_wep(settings):
    settings.ensure_valid()

    if not settings.security_type_is_wep:
        raise RuntimeError("Cannot generate WEP data from non-WEP settings")

    data = {}
    data['submit-url'] = '/wlwep.asp'
    data['submit-value'] = '0.031'
    data['save'] = 'Save'
    data['autogen_key'] = ''
    data['actionType'] = 0
    # I think this field has the side effect
    # of enabling WEP, if it was disabled before..
    data['wepEnabled0'] = 'ON'
    is_64 = settings.security_type == settings.__class__.SECURITY_TYPE_WEP64
    data['length0'] = 1 if is_64 else 2
    data['format0'] = 2 if settings.is_wep_password_in_hex else 1
    data['defaultTxKeyId0'] = 1 # which WEP key is the default
    data['key10'] = settings.password
    data['key20'] = ''
    data['key30'] = ''
    data['key40'] = ''
    return data


def _generate_wireless_data_security(settings):
    settings.ensure_valid()

    data = {}
    data['submit-url'] = '/wlwpa.asp'
    data['submit-value'] = '0.031'
    data['save'] = 'Save'
    if settings.security_type_is_wep:
        data['method0'] = 1
    elif settings.security_type_is_wpa:
        # specifies whether we're using Personal or Radius auth
        data['wpaAuth0'] = 'psk'
        data['pskFormat0'] = '0' # 0 = passphrase; 1 = hex passphrase only
        data['pskValue0'] = settings.password
        data['lifeTime0'] = 86400
        if settings.security_type == settings.__class__.SECURITY_TYPE_WPA:
            data['method0'] = 2
            data['ciphersuite0'] = 'tkip'
        else:
            data['method0'] = 4
            data['wpa2ciphersuite0'] = 'aes'
    else: # no security
        data['method0'] = 0
    return data


def _generate_dmz_data(settings):
    settings.ensure_valid()
    data = {}
    data['submit-url'] = '/dmz.asp'
    data['save'] = 'Save'
    if settings.is_enabled:
        data['enabled'] = 'ON'
        data['ip'] = settings.ip
    return data


def _generate_addr_reservation_item_data(item):
    data = {}
    data['addLease'] = 'Save'
    data['leaseIp'] = item.ip
    data['mac'] = item.mac
    data['submit-url'] = '/tcpiplan.asp'
    data['submit-value'] = '0.031'
    return data



import urllib
import urllib2
import contextlib

from roscraco.exception import RouterFetchError


class RouterBase(object):
    """Base router controller class.

    All other controller classes inherit from this one,
    (hopefully) implementing all of its non-implemented methods.
    """

    def __init__(self, host, port, username, password):
        self._host = host
        self._port = int(port)
        self._username = username
        self._password = password

        self._requests_cache = {}
        self._is_first_request = True

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port

    @property
    def username(self):
        return self._username

    @property
    def password(self):
        return self._password

    @property
    def url_base(self):
        """The base URL to which all requests would be relative to.

        Subclasses may wish to override this to make things
        more convenient.
        """
        return 'http://%s:%d/' % (self.host, self.port)

    def __not_implemented(self):
        raise NotImplementedError('Method needs to be implemented in a subclass.')

    def get_router_info(self):
        """Returns a :class:`roscraco.response.RouterInfo` object
        with system information about the router (hardware, firmware) version.
        """
        self.__not_implemented()

    def get_mac_address(self):
        """Returns the MAC address of the router in our so called
        generic format - lowercase HEX, no separators.
        """
        self.__not_implemented()

    def get_dns_servers(self):
        """Returns a list of IP addresses of the detected DNS servers."""
        self.__not_implemented()

    def get_pppoe_online_time(self):
        """Returns the time in seconds since the PPPoE connection
        was made.

        Returns None if such information is not available
        """
        self.__not_implemented()

    def get_uptime(self):
        """Returns the time in seconds since the router was last rebooted."""
        self.__not_implemented()

    def get_traffic_stats(self):
        """Returns a :class:`roscraco.response.TrafficStats` object
        with traffic statistics for the WAN interface.
        """
        self.__not_implemented()

    def get_connected_clients_list(self):
        """Returns a :class:`roscraco.response.ConnectedClientsList` object
        with information about clients that are current connected to the
        router, or have been connected to it recently.
        """
        self.__not_implemented()

    def get_dmz_settings(self):
        """Returns a :class:`roscraco.response.DMZSettings` object
        with DMZ settings information about the router.
        """
        self.__not_implemented()

    def push_dmz_settings(self, settings):
        """Updates the DMZ settings and returns True on success.

        :param settings: the DMZSettings object
        """
        self.__not_implemented()

    def get_dhcp_settings(self):
        """Returns a :class:`roscraco.response.DHCPServerSettings` object
        with information about the current status of the DHCP server
        (on/off, supported/non-supported, IP range, etc.).
        """
        self.__not_implemented()

    def get_addr_reservation_list(self):
        """Returns a :class:`roscraco.response.DHCPReservationList`
        with information about the current static entries in the DHCP
        reservation list.
        """
        self.__not_implemented()

    def push_addr_reservation_list(self, lst_new):
        """Updates the DHCP address reservation settings and returns True
        on success.

        :param lst_new: the DHCPReservationList object
        """
        self.__not_implemented()

    def get_wireless_settings(self):
        """Returns a :class:`roscraco.response.WirelessSettings` object
        with information about the current status of wireless connectivity
        for this router.
        """
        self.__not_implemented()

    def push_wireless_settings(self, settings):
        """Updates the wireless settings and returns True on success.

        :param settings: the WirelessSettings object
        """
        self.__not_implemented()

    @property
    def supports_reboot(self):
        """Tells whether the device can be rebooted from the web interface."""
        self.__not_implemented()

    def reboot(self):
        """Reboots the device."""
        self.__not_implemented()

    def confirm_identity(self):
        """Confirms that the router we're trying to access
        is of the expected type/model.

        If it's not it will throw an exception of type
        :class:`roscraco.exception.RouterIdentityError`.
        """
        self.__not_implemented()

    def invalidate_cache(self):
        """Clears any cached data that may have been gathered
        and starts clean.
        """
        self._invalidate_http_cache()

    def _invalidate_http_cache(self):
        """Clears the http requests cache."""
        self._requests_cache = {}

    def _make_http_request_read(self, path):
        """Makes an HTTP request to the given path.

        The path is prefixed with the base URL,
        obtained from :attr:`roscraco.router.RouterBase.url_base`.

        Read requests to the same URL are cached, until a write
        request comes along, which invalidates the cache.
        """
        url = self.url_base + path
        if url not in self._requests_cache:
            self._requests_cache[url] = self._perform_http_request(url)[2]
        return self._requests_cache[url]

    def _make_http_request_write(self, path, *args, **kwargs):
        """Makes a "write" HTTP request to the given path.

        ``Write`` here means that the request is potentially
        changing some data, so we're invalidating the http
        requests cache.

        Such requests sometimes use the POST request method,
        although not necessarily.
        """
        url = self.url_base + path
        self._invalidate_http_cache()
        return self._perform_http_request(url, *args, **kwargs)[2]

    def _handle_first_request(self):
        """To be implemented in a subclass
        (if it needs to do something special on the first request).
        """
        pass

    def _perform_http_request(self, url, data=None, headers=(), timeout=7.0):
        """Makes the actual HTTP request and returns the result.

        The result is a 3-tuple:
            - requested URL
            - info() - meta information, such as headers
            - contents
        """
        if self._is_first_request:
            self._is_first_request = False
            self._handle_first_request()

        if data is not None:
            if isinstance(data, dict) or isinstance(data, list):
                data = urllib.urlencode(data)
            else:
                raise RouterFetchError(
                    'POST data should be a dict, a list or None!'
                )

        try:
            req = urllib2.Request(url, data)
            for header, value in headers:
                req.add_header(header, value)
            with contextlib.closing(urllib2.urlopen(req, timeout=timeout)) as handle:
                self._is_logged_in = True
                return (
                    handle.geturl(),
                    handle.info(),
                    handle.read().decode('utf-8', 'ignore')
                )
        except Exception, e:
            raise RouterFetchError('Failed making request: %s' % repr(e))

    def close(self):
        """Performs cleanup, logout or whatever the device requires."""

    def __del__(self):
        try:
            self.close()
        except Exception:
            # close() may fail if __init__ didn't complete
            pass



class RouterError(Exception):
    pass

class RouterFetchError(RouterError):
    pass

class RouterParseError(RouterError):
    pass

class RouterLoginError(RouterError):
    pass

class RouterSettingsError(RouterError):
    pass

class RouterIdentityError(RouterError):
    pass

class RouterNotSupported(RouterError):
    pass



from roscraco.helper import validator, converter
from roscraco.exception import RouterSettingsError


class DHCPReservationList(list):
    """A response class representing the router's
    address reservation list (static DHCP entries).
    """

    def __init__(self, *args, **kwargs):
        list.__init__(self, *args, **kwargs)
        self._supports_reservations = True
        self._changes_require_reboot = True
        self._internal_params = {}

    def set_reservation_support_status(self, value):
        self._supports_reservations = bool(value)

    @property
    def supports_reservations(self):
        """Tells whether address reservation is supported
        by the given device.
        """
        return self._supports_reservations

    def set_reboot_requirement_status(self, value):
        self._changes_require_reboot = bool(value)

    @property
    def changes_require_reboot(self):
        """Tells whether changing address reservation settings
        requires a device reboot for changes to take effect.
        """
        return self._changes_require_reboot

    def set_internal_param(self, key, value):
        """Sets some internal information to be 'carried' with this instance.

        This information may be needed when pushing new settings.
        """
        self._internal_params[key] = value

    def get_internal_param(self, k):
        """Retrieves the internal information
        set by :meth:`set_internal_param`.
        """
        return self._internal_params[k] if k in self._internal_params else None

    def has_ip(self, ip):
        """Tells whether the given IP address has an entry
        in the DHCP reservation list.
        """
        for item in self:
            if item.ip == ip:
                return True
        return False

    def has_ip_mac_entry(self, ip, mac):
        """Tells whether there's an entry that matches both
        the given IP and MAC address.
        """
        for item in self:
            if item.ip == ip and item.mac == mac:
                return True
        return False

    def validate(self):
        for item in self:
            errors = item.validate()
            if len(errors) != 0:
                return errors
        return {}

    def ensure_valid(self):
        for item in self:
            item.ensure_valid()

    def __eq__(self, other):
        if len(self) != len(other):
            return False

        for idx, item in enumerate(self):
            if item != other[idx]:
                return False
        return True

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return id(self)

    def export(self):
        """Exports the most important settings attributes,
        omitting any internal attributes.
        """
        export = {}
        for i, item in enumerate(self):
            export['entry_%d' % i] = item.export()
        return export


class DHCPReservationListItem(object):
    """Represents a single entry in the DHCP address reservation table."""

    #: Properties to export()
    PROPERTIES = (
        'mac', 'ip'
    )

    def __init__(self):
        self._mac_address = None
        self._ip_address = None
        self._is_enabled = True

    def set_mac(self, value):
        self._mac_address = converter.normalize_mac(value)

    @property
    def mac(self):
        return self._mac_address

    def set_ip(self, value):
        self._ip_address = value

    @property
    def ip(self):
        return self._ip_address

    def set_enabled_status(self, value):
        self._is_enabled = bool(value)

    @property
    def is_enabled(self):
        return self._is_enabled

    def validate(self):
        errors = {}
        if not validator.is_valid_ip_address(self.ip):
            errors['ip'] = 'Invalid IP address: %s' % self.ip
        if not validator.is_valid_mac_address_normalized(self.mac):
            errors['mac'] = 'Invalid normalized MAC address: %s' % self.mac
        return errors

    def ensure_valid(self):
        errors = self.validate()
        if len(errors) != 0:
            raise RouterSettingsError(str(errors))

    def __repr__(self):
        return '<%s: %s; %s>' % (self.__class__, self._ip_address, self._mac_address)

    def __eq__(self, other):
        for attr in ('ip', 'mac', 'is_enabled'):
            if getattr(self, attr, None) != getattr(other, attr, None):
                return False
        return True

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return id(self)

    def export(self):
        """Exports the most important settings attributes,
        omitting any internal attributes.
        """
        export = {}
        for attr in self.__class__.PROPERTIES:
            export[attr] = getattr(self, attr, None)
        return export


class DHCPServerSettings(object):
    """Represents the DHCP server settings.

    This is important, because it gives us information on the IP
    range in which addresses can be reserved
    (if DHCP is supported/enabled at all).
    """

    def __init__(self):
        self._ip_start = '192.168.1.100'
        self._ip_end = '192.168.1.199'
        self._is_enabled = True

    @property
    def ip_start(self):
        return self._ip_start

    def set_ip_start(self, value):
        self._ip_start = value

    @property
    def ip_end(self):
        return self._ip_end

    def set_ip_end(self, value):
        self._ip_end = value

    @property
    def is_enabled(self):
        return self._is_enabled

    def set_enabled_status(self, value):
        self._is_enabled = bool(value)

    def validate(self):
        errors = {}

        if not validator.is_valid_ip_address(self._ip_start):
            errors['ip_start'] = 'Invalid start IP: %s' % self._ip_start

        if not validator.is_valid_ip_address(self._ip_end):
            errors['ip_end'] = 'Invalid end IP: %s' % self._ip_end

        if converter.ip2long(self._ip_start) > converter.ip2long(self._ip_end):
            errors['ip_range'] = 'Invalid IP range: from %s to %s' % (self._ip_start, self._ip_end)

        return errors

    def ensure_valid(self):
        errors = self.validate()
        if len(errors) != 0:
            raise RouterSettingsError(str(errors))



# -*- coding: utf-8 -*

"""
    roscraco
    ~~~~~~~~

    A library for managing home routers (networking equipment).

    :copyright: (c) 2011 by Slavi Pantaleev.
    :license: BSD, see LICENSE.txt for more details.
"""

from . import helper
from . import response
from . import router
from exception import RouterNotSupported


__title__ = 'roscraco'
__version__ = '0.3.2'
__author__ = 'Slavi Pantaleev'
__license__ = 'BSD'
__copyright__ = 'Copyright 2011 Slavi Pantaleev'


ROUTER_TP_LINK = 'tplink'
ROUTER_CANYON = 'canyon'
ROUTER_NETGEAR = 'netgear'
ROUTER_TOMATO = 'tomato'
ROUTER_ZYXEL = 'zyxel'
ROUTER_TENDA = 'tenda'


def get_supported_types():
    """Returns a list of supported router types."""
    import sys
    this = sys.modules[__name__]
    return [getattr(this, k) for k in dir(this) if k.startswith('ROUTER_')]


def _get_type_module(router_type):
    """Returns the module that provides the implementation
    for this router type and throws an exception for invalid types."""
    module = getattr(router, router_type, None)
    if module is None:
        raise RouterNotSupported('%s is not a supported router type.' % router_type)
    return module


def _ucfirst(string):
    return string[0].upper() + string[1:]


def get_supported_models(router_type):
    """Returns a list of supported models for the given type
    and throws an exception for invalid types."""
    module = _get_type_module(router_type)
    # All model classes are prefixed this way..
    cls_prefix = _ucfirst(router_type) + '_'
    prefix_len = len(cls_prefix)
    return [k[prefix_len:] for k in dir(module) if k.startswith(cls_prefix)]


def create_controller(router_type, router_model, *args, **kwargs):
    module = _get_type_module(router_type)

    # type=tomato, model=1.23 => class_name=Tomato_1_23
    router_class_name = ''.join((
        _ucfirst(router_type),
        '_',
        router_model.replace('.', '_'),
    ))

    cls = getattr(module, router_class_name, None)
    if cls is None:
        raise RouterNotSupported(
            '%s is not a supported model for %s.' % (
                    router_model,
                    router_type,
                )
        )
    return cls(*args, **kwargs)



import urllib

from roscraco.response import WirelessSettings
from base import TplinkBase, _extract_js_array_data
from roscraco.exception import RouterFetchError

from wr740n import Tplink_WR740N


# Tp-Link WR340G v4 devices seem to be managed just like WR740N devices.
class Tplink_WR340Gv4(Tplink_WR740N):
    def confirm_identity(self):
        self._ensure_www_auth_header('Basic realm="TP-LINK Wireless G Router WR340G"')


class Tplink_WR340G(TplinkBase):

    def get_wireless_settings(self):
        html_settings = self._make_http_request_read('WlanNetworkRpm.htm')
        return _parse_wireless_settings_WR340G(html_settings)

    def push_wireless_settings(self, settings):
        get_params_settings = _generate_wireless_settings_data(settings)
        uri = 'WlanNetworkRpm.htm?%s' % urllib.urlencode(get_params_settings)
        try:
            contents = self._make_http_request_write(uri)
            # Settings are successfully pushed and the router will
            # start the rebooting process if we find this string.
            return 'Please wait a moment' in contents
        except RouterFetchError, e:
            # It sometimes updates the settings and
            # starts rebooting without sending a response correctly.
            # Try to detect that timeout and consider it a success too,
            # even though it may be caused by other reasons.
            if 'timed out' in str(e):
                return True
            raise

    def confirm_identity(self):
        self._ensure_www_auth_header('Basic realm="TP-LINK Wireless Router WR340G"')


def _parse_wireless_settings_WR340G(html_settings):
    """Extracts the Wireless settings from the page contents for a WR340G router."""

    obj = WirelessSettings()
    obj.add_security_support(WirelessSettings.SECURITY_TYPE_WEP64)
    obj.add_security_support(WirelessSettings.SECURITY_TYPE_WEP128)
    obj.add_security_support(WirelessSettings.SECURITY_TYPE_WPA)
    obj.add_security_support(WirelessSettings.SECURITY_TYPE_WPA2)

    settings_array = _extract_js_array_data(html_settings, 'wlanPara')
    wlan_list_array = _extract_js_array_data(html_settings, 'wlanList')

    security_type = int(settings_array[18])
    if security_type == 1: # WEP of some sort
        bit_length = int(wlan_list_array[1])
        if bit_length == 13: # 128 bit length
            security_type = WirelessSettings.SECURITY_TYPE_WEP128
        else:
            security_type = WirelessSettings.SECURITY_TYPE_WEP64
    elif security_type == 3: # WPA-PSK (WPA or WPA2)
        # string like '331', '332', '333'
        # we're interested in the 3rd char, which deals with WPA-PSK/WPA2-PSK
        # 3rd char possible values {1: WPA, 2: WPA2, 3: Automatic (WPA + WPA2)}
        security_options = settings_array[19]

        if int(security_options[2]) == 1:
            security_type = WirelessSettings.SECURITY_TYPE_WPA
        else:
            security_type = WirelessSettings.SECURITY_TYPE_WPA2
    else: # type is either 0 (no security) or 2 (WPA-Enterprise)
        security_type = WirelessSettings.SECURITY_TYPE_NONE
    obj.set_security_type(security_type)

    password = wlan_list_array[0] if obj.security_type_is_wep else settings_array[26]
    obj.set_password(password)

    obj.set_ssid(settings_array[2])
    obj.set_enabled_status(settings_array[8] != 0)
    obj.set_ssid_broadcast_status(settings_array[9] == 1)
    # We don't need to reboot manually..
    # The router reboots by itself when settings are pushed.
    obj.set_reboot_requirement_status(False)
    obj.set_channel(settings_array[6])
    obj.set_internal_param('region', settings_array[4])
    obj.set_internal_param('mode', settings_array[7])

    return obj


def _generate_wireless_settings_data(settings):
    """Generates a wireless settings data array to push to
    the router from the given wireless settings."""

    WirelessSettings = settings.__class__

    settings.ensure_valid()

    data = {}
    data['ssid'] = settings.ssid
    data['channel'] = settings.channel
    data['Save'] = 'Save'

    # preserve some of the params we don't handle
    for k in ('region', 'mode'):
        data[k] = settings.get_internal_param(k)

    if settings.is_enabled:
        data['ap'] = 2 # stands for Active/True

    if settings.is_broadcasting_ssid:
        data['broadcast'] = 2 # stands for Active/True

    if settings.security_type_is_wep:
        data['secType'] = 1
        data['secOpt'] = 3 # Security Option = Automatic
        data['keynum'] = 1 # which of the 4 WEP keys to use (if WEP is enabled)

        # These are the 4 WEP keys and their lengths.. empty by default - to be filled later if needed
        # only one of these keys is actively used (specified by `keynum` above)
        merge_with =  {'key2': '', 'key3': '', 'key4': '', 'length2': 0, 'length3': 0, 'length4': 0}
        data = dict(data, **merge_with)

        # 64bit = 5, 128bit = 13
        bit_length = 5 if settings.security_type == WirelessSettings.SECURITY_TYPE_WEP64 else 13

        data['key1'] = settings.password
        data['length1'] = bit_length
        data['keytype'] = 1 if settings.is_wep_password_in_hex else 2 # ASCII or HEX
    elif settings.security_type_is_wpa:
        data['secType'] = 3
        data['encrptType'] = 1 # Automatic encryption
        data['interval'] = 0 # group update interval
        data['pskSecret'] = settings.password

        if settings.security_type == WirelessSettings.SECURITY_TYPE_WPA:
            data['secOpt'] = 1 # WPA-PSK
        else:
            data['secOpt'] = 2 #WPA2-PSK
    else:
        # security is being disabled
        data['secType'] = 1

        # it's as if we're disabling it while on the WEP settings page..
        # let's send some of the WEP fields

        data['keytype'] = 1

        merge_with =  {'key1': '', 'key2': '', 'key3': '', 'key4': '', 'length1': 0, 'length2': 0, 'length3': 0, 'length4': 0}
        data = dict(data, **merge_with)

    # secStatus != 2 means 'disable security'
    if settings.security_type != WirelessSettings.SECURITY_TYPE_NONE:
        data['secStatus'] = 2 # stands for Active/True

    return data



import re
import ast
import base64

from roscraco.router.base import RouterBase

from roscraco.helper import validator, converter
from roscraco.exception import RouterParseError, \
     RouterSettingsError, RouterIdentityError

from roscraco.response import RouterInfo, TrafficStats, DMZSettings, \
     ConnectedClientsListItem, ConnectedClientsList, WirelessSettings, \
     DHCPReservationList, DHCPReservationListItem, DHCPServerSettings


class Tomato_1_23(RouterBase):

    def _perform_http_request(self, *args, **kwargs):
        auth = base64.b64encode('%s:%s' % (self.username, self.password))
        kwargs['headers'] = [('Authorization', 'Basic %s' % auth)]
        return RouterBase._perform_http_request(self, *args, **kwargs)

    def confirm_identity(self):
        html = self._make_http_request_read('')
        if 'Tomato' not in html:
            raise RouterIdentityError(
                'Cannot confirm that this is a Tomato router!'
            )
        info = self.get_router_info()
        if info.firmware_version != 'Version 1.23':
            raise RouterIdentityError('Tomato router version mismatch!')

    def _get_http_id(self):
        """Extracts the session id (http id),
        which needs to be passed in some requests."""
        html = self._make_http_request_read('')
        return _parse_http_id(html)

    def _get_status_data_raw(self):
        http_id = self._get_http_id()
        status_uri = 'status-data.jsx?_http_id=%s' % http_id
        return self._make_http_request_read(status_uri)

    def get_router_info(self):
        html = self._make_http_request_read('')
        return _parse_router_info(html)

    def get_uptime(self):
        js = self._get_status_data_raw()
        return _parse_uptime(js)

    def get_pppoe_online_time(self):
        js = self._get_status_data_raw()
        return _parse_pppoe_online_time(js)

    def get_traffic_stats(self):
        html = self._make_http_request_read('bwm-monthly.asp')
        return _parse_traffic_stats(html)

    def get_mac_address(self):
        js = self._get_status_data_raw()
        return _parse_mac_address(js)

    def get_dns_servers(self):
        js = self._get_status_data_raw()
        return _parse_dns_servers(js)

    def get_connected_clients_list(self):
        html = self._make_http_request_read('status-devices.asp')
        return _parse_connected_clients_list(html)

    def get_dmz_settings(self):
        html = self._make_http_request_read('forward-dmz.asp')
        return _parse_dmz_settings(html)

    def push_dmz_settings(self, settings):
        http_id = self._get_http_id()
        data = _generate_dmz_data(http_id, settings)
        self._make_http_request_write('tomato.cgi', data)
        return self.get_dmz_settings() == settings

    def get_dhcp_settings(self):
        html = self._make_http_request_read('basic-network.asp')
        return _parse_dhcp_settings(html)

    def get_addr_reservation_list(self):
        html = self._make_http_request_read('basic-static.asp')
        return _parse_addr_reservation_list(html)

    def push_addr_reservation_list(self, lst_new):
        http_id = self._get_http_id()
        data = _generate_addr_reservation_data(http_id, lst_new)
        self._make_http_request_write('tomato.cgi', data)
        return self.get_addr_reservation_list() == lst_new

    def get_wireless_settings(self):
        html = self._make_http_request_read('basic-network.asp')
        return _parse_wireless_settings(html)

    def push_wireless_settings(self, settings):
        http_id = self._get_http_id()
        data = _generate_wireless_data(http_id, settings)
        self._make_http_request_write('tomato.cgi', data)
        return self.get_wireless_settings() == settings

    @property
    def supports_reboot(self):
        return True

    def reboot(self):
        http_id = self._get_http_id()
        data = {'_commit': 0, '_http_id': http_id, '_nvset': 0, '_reboot': 1}
        self._make_http_request_write('tomato.cgi', data)


def _parse_http_id(html):
    regex = "src='status-data.jsx\?_http_id=(TID(?:.+?))'>"
    match_object = re.compile(regex).search(html)
    if match_object is None:
        raise RouterParseError('Cannot determine http id!')
    return match_object.group(1)


def _parse_data_structure(inner_json):
    """Parses broken JSON that this router generates.

    Data structures look like this:
    {
        key: 'value',
        another_key: [1, 2, 3]
    }

    We're basically using what's between the main {},
    and are parsing that, returning a dictionary.
    """

    results = {}
    for line in inner_json.split('\n'):
        match_object = re.compile('^\s+(.+?):\s(.+?),?$').match(line)
        if match_object is None:
                continue
        key, value = match_object.groups()
        try:
            results[key] = ast.literal_eval(value)
        except ValueError:
            pass
    return results


def _parse_js_structure(js, name):
    """Extracts a javascript object by its name.

    Here's what it looks like initially:
    //
    object_name = {
        some stuff here
    };

    //

    We're parsing the broken JSON inside, and
    we're returning a dictionary.
    """

    regex = '//\n%s = \{(.+?)\};\n\n//' % name
    match_object = re.compile(regex, re.DOTALL).search(js)
    if match_object is None:
        raise RouterParseError('Cannot parse main structure data')
    return _parse_data_structure(match_object.group(1))


def _parse_router_info(html):
    regex = re.compile("<div class='version'>(.+?)</div>")
    match_object = regex.search(html)
    if match_object is None:
        raise RouterParseError('Cannot parse version')
    obj = RouterInfo()
    obj.set_hardware_version('Unknown')
    obj.set_firmware_version(match_object.group(1))
    return obj


def _parse_uptime(status_js):
    data = _parse_js_structure(status_js, 'sysinfo')
    try:
        return int(data['uptime'])
    except (KeyError, ValueError):
        raise RouterParseError('Cannot parse uptime!')


def _parse_pppoe_online_time(status_js):
    # Ensure that we're using PPPoE
    nvram = _parse_js_structure(status_js, 'nvram')
    try:
        if nvram['wan_proto'] != 'pppoe':
            return None
        regex = "stats.wanuptime = '(\d+) days?, (\d+):(\d+):(\d+)';"
        match_object = re.compile(regex).search(status_js)
        if match_object is None:
            return None
        days, hours, minutes, seconds = map(int, match_object.groups())
        return days * 86400 + hours * 3600 + minutes * 60 + seconds
    except (KeyError, ValueError):
        raise RouterParseError('Cannot parse pppoe online time!')


def _parse_traffic_stats(html):
    # The monthly_history array contains one or more other arrays.
    # Each of those arrays contains 3 items (like a 3-tuple), which are:
    #   1. time data, for which 2. and 3. are
    #   2. traffic received in KB
    #   3. traffic sent in KB
    regex = 'monthly_history = \[\n(.+?)\];'
    match_object = re.compile(regex).search(html)
    if match_object is not None:
        try:
            array = ast.literal_eval('[%s]' % match_object.group(1))
            kb_recv, kb_sent = 0, 0
            for time_data, recv, sent in array:
                kb_recv += recv
                kb_sent += sent
            return TrafficStats(kb_recv * 1024, kb_sent * 1024, 0, 0)
        except (KeyError, ValueError):
            raise RouterParseError('Cannot parse traffic stats!')
    return TrafficStats(0, 0, 0, 0)


def _parse_mac_address(status_js):
    nvram = _parse_js_structure(status_js, 'nvram')
    try:
        return converter.normalize_mac(nvram['wan_hwaddr'])
    except KeyError:
        raise RouterParseError('Cannot parse MAC address!')


def _parse_dns_servers(status_js):
    regex = re.compile('//\ndns = \[(.+?)\];')
    match_object = regex.search(status_js)
    if match_object is None:
        return []
    try:
        ips = ast.literal_eval(match_object.group(1))
    except Exception:
        return []
    else:
        return [ip.strip(' ') for ip in ips if validator.is_valid_ip_address(ip)]


def _parse_connected_clients_list(html):
    lst = ConnectedClientsList()

    regex = re.compile('dhcpd_lease = \[ (.+?)\];\nlist = \[')
    match_object = regex.search(html)
    if match_object is None:
        return lst

    def parse_lease_time(lease):
        regex = '(\d+) days?, (\d+):(\d+):(\d+)'
        match_object = re.compile(regex).search(lease)
        if match_object is None:
            return ConnectedClientsListItem.__class__.LEASE_TIME_PERMANENT

        days, hours, minutes, seconds = map(int, match_object.groups())
        return days * 86400 + hours * 3600 + minutes * 60 + seconds

    try:
        array = ast.literal_eval('[%s]' % match_object.group(1))
        for name, ip, mac, lease_time in array:
            item = ConnectedClientsListItem()
            item.set_client_name(name)
            item.set_ip(ip)
            item.set_mac(converter.normalize_mac(mac))
            item.set_lease_time(parse_lease_time(lease_time))
            lst.append(item)
        return lst
    except ValueError:
        raise RouterParseError('Cannot parse connected clients!')


def _parse_dmz_settings(html):
    regex = "//\s+nvram = \{(.+?)\};\n\nvar lipp = '(.+?)';"
    match_object = re.compile(regex, re.DOTALL).search(html)
    if match_object is None:
        raise RouterParseError('Cannot parse DMZ settings')

    bad_json_settings, lipp = match_object.groups()
    nvram = _parse_data_structure(bad_json_settings)

    ip = nvram['dmz_ipaddr']
    if '.' not in ip:
        # it's the last part only.. it's shortened
        # and the rest is in lipp
        ip = lipp + ip

    obj = DMZSettings()
    obj.set_supported_status(True)
    obj.set_reboot_requirement_status(False)
    obj.set_enabled_status(nvram['dmz_enable'] == '1')
    obj.set_ip(ip)
    obj.ensure_valid()
    return obj


def _parse_dhcp_settings(html):
    regex = '//\s+nvram = \{(.+?)\};\n\nxob = '
    match_object = re.compile(regex, re.DOTALL).search(html)
    if match_object is None:
        raise RouterParseError('Cannot parse DHCP settings')

    array = _parse_data_structure(match_object.group(1))

    try:
        ip_start, ip_end = array['dhcpd_startip'], array['dhcpd_endip']
        is_enabled = array['lan_proto'] == 'dhcp'
    except KeyError:
        raise RouterParseError('Bad nvram for DHCP settings')

    if not validator.is_valid_ip_address(ip_start):
        raise RouterParseError('Invalid DHCP start IP: %s' % ip_start)

    if not validator.is_valid_ip_address(ip_end):
        raise RouterParseError('Invalid DHCP end IP: %s' % ip_end)

    obj = DHCPServerSettings()
    obj.set_enabled_status(is_enabled)
    obj.set_ip_start(ip_start)
    obj.set_ip_end(ip_end)
    obj.ensure_valid()
    return obj


def _parse_addr_reservation_list(html):
    regex = '//\s+nvram = \{(.+?)\};\n\nif '
    match_object = re.compile(regex, re.DOTALL).search(html)
    if match_object is None:
        raise RouterParseError('Cannot parse reservation list')

    array = _parse_data_structure(match_object.group(1))
    try:
        lst = DHCPReservationList()
        lst.set_reboot_requirement_status(False)

        for part in array['dhcpd_static'].split('>'):
            if part == '':
                continue
            mac, ip, name = part.split('<')
            item = DHCPReservationListItem()
            item.set_mac(mac)
            item.set_ip(ip)
            lst.append(item)
    except (KeyError, ValueError):
        raise RouterParseError('Bad nvram for reservation list')

    return lst


def _parse_wireless_settings(html):
    regex = '//\s+nvram = \{(.+?)\};\n\nxob = '
    match_object = re.compile(regex, re.DOTALL).search(html)
    if match_object is None:
        raise RouterParseError('Cannot parse wireless settings')

    array = _parse_data_structure(match_object.group(1))

    try:
        settings = WirelessSettings()
        settings.add_security_support(WirelessSettings.SECURITY_TYPE_WEP64)
        settings.add_security_support(WirelessSettings.SECURITY_TYPE_WEP128)
        settings.add_security_support(WirelessSettings.SECURITY_TYPE_WPA)
        settings.add_security_support(WirelessSettings.SECURITY_TYPE_WPA2)

        settings.set_reboot_requirement_status(False)
        settings.set_auto_channel_support(False)
        settings.set_ascii_wep_password_support_status(False)

        # Let's preserve all the settings, so that it's
        # easier to generate the data later
        settings.set_internal_param('nvram', array)

        sec_type = array['security_mode2']
        settings.set_security_type(WirelessSettings.SECURITY_TYPE_NONE)
        settings.set_password('')
        if sec_type == 'wep':
            if array['wl_wep_bit'] == '64':
                settings.set_security_type(WirelessSettings.SECURITY_TYPE_WEP64)
            else:
                settings.set_security_type(WirelessSettings.SECURITY_TYPE_WEP128)
            settings.set_password(array['wl_key1'])
        elif sec_type == 'wpa_personal':
            settings.set_security_type(WirelessSettings.SECURITY_TYPE_WPA)
            settings.set_password(array['wl_wpa_psk'])
        elif sec_type == 'wpa2_personal':
            settings.set_security_type(WirelessSettings.SECURITY_TYPE_WPA2)
            settings.set_password(array['wl_wpa_psk'])

        settings.set_ssid(array['wl_ssid'])
        settings.set_channel(array['wl_channel'])
        settings.set_ssid_broadcast_status(array['wl_closed'] == '0')
        settings.set_enabled_status(array['wl_radio'] == '1')
        return settings
    except (KeyError, ValueError):
        raise RouterParseError('Bad nvram for wireless settings')


def _generate_addr_reservation_data(http_id, lst_new):
    lst_new.validate()

    items = []
    for i, item in enumerate(lst_new):
        name = 'Client%d' % i
        items.append('<'.join((item.mac, item.ip, name)))

    if len(items) == 0:
        items_string = ''
    else:
        items.append('')
        items_string = '>'.join(items)

    data = {}
    data['ajax'] = 1
    data['_nextpage'] = 'basic-static.asp'
    data['_service'] = 'dhcpd-restart'
    data['_http_id'] = http_id
    data['dhcpd_static'] = items_string
    return data


def _generate_dmz_data(http_id, settings):
    settings.ensure_valid()

    data = {}
    data['_ajax'] = 1
    data['_http_id'] = http_id
    data['dmz_sip'] = ''
    data['_nextpage'] = 'forward-dmz.asp'
    data['_service'] = 'firewall-restart'
    data['dmz_enable'] = 1 if settings.is_enabled else 0
    data['dmz_ipaddr'] = settings.ip
    return data


def _generate_wireless_data(http_id, settings):
    settings.ensure_valid()

    nvram = settings.get_internal_param('nvram')
    if nvram is None:
        raise RouterSettingsError('Bad wireless settings. Missing nvram')

    # We'll use all the settings (nvram) as a base for the new stuff
    # Certain fields are 'status fields', so we
    # don't need to push them as settings
    #
    # Other fields (like the WPA password one) may be empty if when using WEP,
    # so pushing an empty WPA password field would be invalid
    attrs_skip = (
        'dhcp_num', 'dhcp_start', 'lan_gateway',
        'wl_radius_port', 'wl0_hwaddr', 'http_id',
        'wan_ipaddr', 'pptp_server_ip', 'l2tp_server_ip',
        'wan_netmask', 'wan_gateway', 'ppp_idletime',
        'wl_radius_ipaddr', 'wl_radius_key',
        'wl_wpa_psk', 'wl_key1'
    )

    data = {}
    for k, v in nvram.items():
        if k not in attrs_skip:
            data[k] = v

    data['_http_id'] = http_id
    data['_ajax'] = 1
    data['_service'] = '*'
    data['_nextpage'] = 'basic-network.asp'
    data['_nextwait'] = 10
    data['wl_gmode'] = 1 # 0: B; 1: B+G; 2: B

    data['wl_radio'] = 1 if settings.is_enabled else 0
    data['wl_channel'] = settings.channel
    data['wl_ssid'] = settings.ssid
    data['wl_closed'] = 0 if settings.is_broadcasting_ssid else 1

    # `none` or `radius`
    data['wl_auth_mode'] = 'none'

    data['wl_wep'] = 'enabled' if settings.security_type_is_wep else 'disabled'

    # should be changed to 1 if we're changing the LAN IP
    data['_moveip'] = 0

    # value is good for security type None and WEP
    data['wl_akm'] = ''

    if settings.security_type_is_wpa:
        if settings.security_type == settings.__class__.SECURITY_TYPE_WPA:
            data['wl_akm'] = 'psk'
        else:
            data['wl_akm'] = 'psk2'


    data['security_mode'] = 'disabled'
    data['security_mode2'] = 'disabled'
    if settings.security_type_is_wep:
        data['security_mode'] = 'wep'
        data['security_mode2'] = 'wep'
        if settings.security_type == settings.__class__.SECURITY_TYPE_WEP64:
            data['wl_wep_bit'] = 64
        else:
            data['wl_wep_bit'] = 128
        data['wl_key'] = 1
        data['wl_key1'] = settings.password
    elif settings.security_type_is_wpa:
        data['wl_wpa_psk'] = settings.password
        data['wl_crypto'] = 'tkip+aes'
        if settings.security_type == settings.__class__.SECURITY_TYPE_WPA:
            data['security_mode'] = 'psk'
            data['security_mode2'] = 'wpa_personal'
        else:
            data['security_mode'] = 'psk2'
            data['security_mode2'] = 'wpa2_personal'

    return data



import re
import base64

from roscraco.router.base import RouterBase

from roscraco.helper import validator, converter
from roscraco.exception import RouterFetchError, RouterParseError, \
     RouterIdentityError

from roscraco.response import RouterInfo, TrafficStats, DMZSettings, \
     ConnectedClientsListItem, ConnectedClientsList, WirelessSettings, \
     DHCPReservationList, DHCPReservationListItem, DHCPServerSettings


class NetgearWGR614Base(RouterBase):
    """Base class for all WGR614 routers, regardless of their version."""

    def __init__(self, *args, **kwargs):
        RouterBase.__init__(self, *args, **kwargs)
        self._is_logged_in = False

    def _perform_http_request(self, *args, **kwargs):
        auth = base64.b64encode('%s:%s' % (self.username, self.password))
        kwargs['headers'] = [('Authorization', 'Basic %s' % auth)]
        return RouterBase._perform_http_request(self, *args, **kwargs)

    def _handle_first_request(self):
        """We need to make an initial "login" request.

        The router will respond with 401 Unauthorized (RouterFetchError)
        to other requests if we start making requests directly.
        What we're doing here is making a request which catches that
        initial 401 error and simply ignores it.

        This function also does something else special for Netgear routers.
        The web management interface can be used from only one IP at a time.
        Access to the router is locked until the user logs out
        (handled in :meth:`close`) or his session expires.
        If the router replies with a non-401 response on our first request,
        it may be telling us that someone else is logged in and managing it.
        """
        try:
            contents = self._make_http_request_read('RST_status.htm')
        except RouterFetchError:
            # it's quite normal if we're here!
            pass
        else:
            # We got a good response, that's somewhat abnormal
            # It could mean that someone else is managing the device
            # Or that we're already "logged in" somehow
            if ') is managing this device' in contents:
                regex = re.compile('\((.+?)\) is managing this device')
                match_obj = regex.search(contents)
                ip = match_obj.group(1) if match_obj is not None else 'unknown'
                raise RouterFetchError('Device is managed by %s!' % ip)

    def close(self):
        # It's a good idea to perform a logout, because no one else
        # will be able to log in until the current session expires
        # (the session is on an IP basis)
        try:
            if self._is_logged_in:
                # Only logout if we've made at least one
                # successful request
                self._make_http_request_read('LGO_logout.htm')
                self._is_logged_in = False
        except RouterFetchError:
            pass

    def get_router_info(self):
        return _parse_router_info(self._make_http_request_read('RST_status.htm'))

    def get_uptime(self):
        return _parse_uptime_to_seconds(self._make_http_request_read('RST_stattbl.htm'))

    def get_traffic_stats(self):
        return _parse_traffic_stats(self._make_http_request_read('RST_stattbl.htm'))

    def get_pppoe_online_time(self):
        return _parse_pppoe_online_time(self._make_http_request_read('RST_st_poe.htm'))

    def get_mac_address(self):
        return _parse_mac_address(self._make_http_request_read('RST_status.htm'))

    def get_dns_servers(self):
        return _parse_dns_servers(self._make_http_request_read('RST_status.htm'))

    def get_connected_clients_list(self):
        return _parse_connected_clients_list(self._make_http_request_read('DEV_device.htm'))

    def _confirm_identity_meta_description(self, value):
        contents = self._perform_http_request('%sRST_status.htm' % self.url_base)[2]

        string_to_find = '<META name="description" content="%s' % value
        if string_to_find not in contents:
            raise RouterIdentityError('Cannot find string in contents: %s' % string_to_find)

    def get_dhcp_settings(self):
        return _parse_dhcp_settings(self._make_http_request_read('LAN_lan.htm'))

    def get_addr_reservation_list(self):
        return _parse_addr_reservation_list(self._make_http_request_read('LAN_lan.htm'))

    def push_addr_reservation_list(self, lst_new):
        lst_new.ensure_valid()

        lst_old = self.get_addr_reservation_list()

        if lst_old == lst_new:
            return True

        submit_token = lst_old.get_internal_param('submit_token')

        # delete all old (currently added) items first
        delete_url = 'lan.cgi' if submit_token is None else 'lan.cgi?id=%d' % submit_token
        for _ in lst_old:
            # delete from the front (id=0) as many times as needed
            data = {'Add': '', 'Edit': '', 'Delete': 'Delete', 'select': 0}
            self._make_http_request_write(delete_url, data)

        add_url = 'reserv.cgi' if submit_token is None else 'reserv.cgi?id=%d' % submit_token
        for id, item in enumerate(lst_new):
            data = _generate_addr_reservation_item_data(item)
            self._make_http_request_write(add_url, data)

        return self.get_addr_reservation_list() == lst_new

    @property
    def supports_reboot(self):
        return False

    def reboot(self):
        pass


class Netgear_WGR614v9(NetgearWGR614Base):

    def confirm_identity(self):
        return self._confirm_identity_meta_description('WGR614V9')

    def get_dmz_settings(self):
        return _parse_dmz_settings_WGR614v9(self._make_http_request_read('WAN_wan.htm'))

    def push_dmz_settings(self, settings):
        data = _generate_dmz_data_WGR614v9(settings)
        self._make_http_request_write('security.cgi', data)
        return True

    def get_wireless_settings(self):
        # Maps a security type with the document id, which contains its settings
        security_types_map = {'None': 3, 'WEP': 1, 'WPA-PSK': 2, 'WPA2-PSK': 2, 'WPA-AUTO-PSK': 2}

        html_initial = self._make_http_request_read('WLG_wireless.htm')
        match_object = re.compile('var secuType="(%s)";' % '|'.join(security_types_map)).search(html_initial)
        if match_object is None:
            raise RouterFetchError('Cannot parse initial wireless settings')
        doc_name = 'WLG_wireless%d.htm' % security_types_map[match_object.group(1)]

        args = [match_object.group(1)]
        args.append(self._make_http_request_read(doc_name))
        args.append(self._make_http_request_read('WLG_adv.htm'))

        return _parse_wireless_settings(*args)

    def push_wireless_settings(self, settings):
        submit_token = settings.get_internal_param('submit_token')

        data_basic = _generate_wireless_basic_settings_WGR614v9(settings)
        path_basic = 'wireless.cgi' if submit_token is None else 'wireless.cgi?id=%d' % submit_token
        self._make_http_request_write(path_basic, data_basic)

        data_advanced = _generate_wireless_advanced_settings_WGR614v9(settings)
        path_advanced = 'wlg_adv.cgi' if submit_token is None else 'wlg_adv.cgi?id=%d' % submit_token
        self._make_http_request_write(path_advanced, data_advanced)

        return self.get_wireless_settings() == settings


class Netgear_WGR614v8(Netgear_WGR614v9):

    def confirm_identity(self):
        return self._confirm_identity_meta_description('WGR614v8')

    def get_dmz_settings(self):
        return _parse_dmz_settings_WGR614v8(self._make_http_request_read('WAN_wan.htm'))

    def push_dmz_settings(self, settings):
        data = _generate_dmz_data_WGR614v8(settings)
        self._make_http_request_write('security.cgi', data)
        return True

    def push_wireless_settings(self, settings):
        data_basic = _generate_wireless_basic_settings_WGR614v9(settings)
        self._make_http_request_write('wireless.cgi', data_basic)

        data_advanced = _generate_wireless_advanced_settings_WGR614v8(settings)
        self._make_http_request_write('wlg_adv.cgi', data_advanced)

        return self.get_wireless_settings() == settings


class Netgear_WGR614v7(Netgear_WGR614v8):

    def confirm_identity(self):
        return self._confirm_identity_meta_description('WGR614v7')

    def push_dmz_settings(self, settings):
        data = _generate_dmz_data_WGR614v7(settings)
        self._make_http_request_write('security.cgi', data)
        return True

    def get_wireless_settings(self):
        html_initial = self._make_http_request_read('WLG_wireless.htm')

        match_object = re.compile('var authType="(.+?)";').search(html_initial)
        if match_object is None:
            raise RouterFetchError('Cannot parse initial wireless settings/authType')
        auth_type = match_object.group(1)

        match_object = re.compile('var wepStatus="(.+?)";').search(html_initial)
        if match_object is None:
            raise RouterFetchError('Cannot parse initial wireless settings/wepStatus')
        wep_status = match_object.group(1)

        if wep_status == 'Enable':
            if auth_type in ('WPA-PSK', 'WPA2-PSK', 'WPA-AUTO-PSK'):
                doc_name = 'WLG_wireless2.htm'
            else:
                auth_type = 'WEP' # I've seen auth_type='None' to mean 'WEP'.. weird..
                doc_name = 'WLG_wireless1.htm'
        else:
            doc_name = 'WLG_wireless3.htm'

        args = [auth_type]
        args.append(self._make_http_request_read(doc_name))
        args.append(self._make_http_request_read('WLG_adv.htm'))

        settings = _parse_wireless_settings(*args)
        # Only v7 doesn't support auto channel
        settings.set_auto_channel_support(False)

        return settings

    def push_wireless_settings(self, settings):
        # It's not really confirmed, but pushing advanced before basic looks better
        # because basic restarts the router shortly after it's being changed
        # i can't say the same about advanced though
        data_advanced = _generate_wireless_advanced_settings_WGR614v7(settings)
        data_basic = _generate_wireless_basic_settings_WGR614v9(settings)

        # Shortly after we push settings, the router reboots automatically
        # so the settings we try to push last may not make it,
        # especially if the connection is laggy
        self._make_http_request_write('wlg_adv.cgi', data_advanced)
        self._make_http_request_write('wireless.cgi', data_basic)

        # the router reboots automatically very soon
        # we can't make sure everything went fine..
        return True


def _parse_router_info(html):
    obj = RouterInfo()

    regex_hardware = re.compile('<b>Hardware Version</b>(?:.+?)>([a-zA-Z0-9]+)</td>', re.DOTALL)
    match_object = regex_hardware.search(html)
    if match_object is None:
        raise RouterParseError('Cannot parse hardware version')
    obj.set_hardware_version(match_object.group(1))

    regex_firmware = re.compile('<b>Firmware Version </b>(?:.+?)>([a-zA-Z0-9\s\.\(\)_]+)</td>', re.DOTALL)
    match_object = regex_firmware.search(html)
    if match_object is None:
        raise RouterParseError('Cannot parse firmware version')
    obj.set_firmware_version(match_object.group(1))

    return obj


def _parse_uptime_to_seconds(html):
    """Parses an uptime string such as ` 15 days 10:11:12` or `10:11:12` to seconds."""

    regex = re.compile('<!>(?: ([0-9]+) days? )?([0-9]+):([0-9]+):([0-9]+)<!>')
    match_object = regex.search(html)
    if match_object is None:
        raise RouterParseError('Cannot match uptime string')

    days, hours, minutes, seconds = map(lambda x: int(x) if x is not None else 0, match_object.groups())

    return days * 86400 + hours * 3600 + minutes * 60 + seconds


def _parse_traffic_stats(html):
    # Sadly, we can only get packets count information.. no size stats available

    regex = re.compile('WAN</span>(?:.+?)([0-9]+)</span>(?:.+?)([0-9]+)</span>', re.DOTALL)
    match_object = regex.search(html)
    if match_object is None:
        raise RouterParseError('Cannot find traffic stats information')

    packets_sent, packets_recv = map(int, match_object.groups())

    return TrafficStats(0, 0, packets_recv, packets_sent)


def _parse_pppoe_online_time(html):
    # <TD NOWRAP width="50%">23:05:47</td>
    # or
    # <TD NOWRAP width="50%"> 1 day 00:41:36</td>
    # or
    # <TD NOWRAP width="50%"> 10 days 00:41:36</td>

    regex = re.compile('<B>Connection Time</B>(?:.+?)>(?: ([0-9]+) days? )?([0-9]+):([0-9]+):([0-9]+)</td>', re.DOTALL)
    match_object = regex.search(html)
    if match_object is None:
        raise RouterParseError('Cannot parse online time information')

    days, hours, minutes, seconds = map(lambda x: int(x) if x is not None else 0, match_object.groups())

    return days * 86400 + hours * 3600 + minutes * 60 + seconds


def _parse_mac_address(html):
    regex = re.compile('Internet Port(?:.+?)MAC Address </b></td>(?:.+?)>(.+?)</td>', re.DOTALL)
    match_object = regex.search(html)
    if match_object is None:
        raise RouterParseError('Cannot parse MAC address information')

    return converter.normalize_mac(match_object.group(1))


def _parse_dns_servers(html):
    regex = re.compile('Domain Name Server <!><br></b></td>(?:.+?)>(.+?)</td>', re.DOTALL)
    match_object = regex.search(html)
    if match_object is None:
        raise RouterParseError('Cannot parse DNS servers information')

    ips = match_object.group(1) #<br> separated list of IPs
    return [ip for ip in ips.split('<br>') if validator.is_valid_ip_address(ip)]


def _parse_connected_clients_list(html):
    lst = ConnectedClientsList()

    regex = "<tr>(?:.+?)<span class=\"ttext\">(.+?)</span>(?:.+?)<span class=\"ttext\">(.+?)</span>(?:.+?)<span class=\"ttext\">(.+?)</span>(?:.+?)</tr>"

    for ip, name, mac in re.compile(regex, re.DOTALL).findall(html):
        if ip == '--':
            # I've seen such entries on WGR614v7 only.. let's ignore them
            continue

        if not validator.is_valid_ip_address(ip):
            raise RouterParseError('Invalid IP address: %s' % ip)

        item = ConnectedClientsListItem()
        item.set_client_name(name)
        item.set_mac(converter.normalize_mac(mac))
        item.set_ip(ip)
        item.set_lease_time(None) # no lease time information available

        lst.append(item)

    return lst


def _parse_dmz_settings_WGR614v9(html):
    match_object = re.compile("name=\"dmz_ip\" type=hidden value= \"(.+?)\">").search(html)
    if match_object is None:
        raise RouterParseError('Cannot parse DMZ IP')
    ip = match_object.group(1)
    if not validator.is_valid_ip_address(ip):
        raise RouterParseError('Invalid DMZ IP address: %s' % ip)


    match_object = re.compile("var dmzEnable = \"([0-1])\";").search(html)
    if match_object is None:
        raise RouterParseError('Cannot parse DMZ Enabled Status')
    is_enabled = True if int(match_object.group(1)) == 1 else False


    settings = DMZSettings()
    settings.set_supported_status(True)
    settings.set_ip(ip)
    settings.set_enabled_status(is_enabled)

    return settings


def _parse_dmz_settings_WGR614v8(html):
    match_object = re.compile('var lanIpAddr = "(.+?)";').search(html)
    if match_object is None:
        raise RouterParseError('Cannot parse DMZ IP')
    ip = match_object.group(1)
    if not validator.is_valid_ip_address(ip):
        raise RouterParseError('Invalid DMZ IP address: %s' % ip)


    match_object = re.compile("var dmzEnable = \"([0-1])\";").search(html)
    if match_object is None:
        raise RouterParseError('Cannot parse DMZ Enabled Status')
    is_enabled = True if int(match_object.group(1)) == 1 else False


    settings = DMZSettings()
    settings.set_supported_status(True)
    settings.set_ip(ip)
    settings.set_enabled_status(is_enabled)

    return settings

def __parse_wep_password(html_basic):
    regex = re.compile('<input type="text" name="KEY1" maxLength=32 size=34 value="(.+?)" onkeydown')
    match_object = regex.search(html_basic)
    if match_object is None:
        raise RouterParseError('Cannot determine WEP password')

    return match_object.group(1)

def _parse_wireless_settings(security_type, html_basic, html_advanced):
    settings = WirelessSettings()
    settings.add_security_support(WirelessSettings.SECURITY_TYPE_WEP64)
    settings.add_security_support(WirelessSettings.SECURITY_TYPE_WEP128)
    settings.add_security_support(WirelessSettings.SECURITY_TYPE_WPA)
    settings.add_security_support(WirelessSettings.SECURITY_TYPE_WPA2)
    settings.set_ascii_wep_password_support_status(False)
    settings.set_reboot_requirement_status(False)

    # Determine the submit token.. some WGR614v9 models have such a token
    # It's either some form of CSRF protection or something else..
    match_object = re.compile('<form method="POST" action="wireless.cgi\?id=([0-9]+)">').search(html_basic)
    if match_object is None:
        settings.set_internal_param('submit_token', None)
    else:
        settings.set_internal_param('submit_token', int(match_object.group(1)))

    if security_type == 'WEP':
        if '<option selected value="1">64bit</option>' in html_basic:
            settings.set_security_type(WirelessSettings.SECURITY_TYPE_WEP64)
        elif '<option selected value="2">128bit</option>' in html_basic:
            settings.set_security_type(WirelessSettings.SECURITY_TYPE_WEP128)
        else:
            raise RouterParseError('Cannot determine WEP key length')

        settings.set_password(__parse_wep_password(html_basic))
    elif security_type == 'WPA-PSK':
        settings.set_security_type(WirelessSettings.SECURITY_TYPE_WPA)
        # password extraction is done below
    elif security_type in ('WPA2-PSK', 'WPA-AUTO-PSK'):
        settings.set_security_type(WirelessSettings.SECURITY_TYPE_WPA2)
        # password extraction is done below
    else: # security_type = `Disable` or something else that we don't handle..
        settings.set_security_type(WirelessSettings.SECURITY_TYPE_NONE)
        settings.set_password('')

    if settings.security_type_is_wpa:
        regex = re.compile('<input type="text" name="passphrase" size=20 maxLength=64 value="(.+?)" onFocus')
        match_object = regex.search(html_basic)
        if match_object is None:
            raise RouterParseError('Cannot determine WPA password')
        password = match_object.group(1)
        if '*****' in password:
            # WGR614v7 doesn't present us with the real password, but substitutes it with * chars
            # that's not the case for v8 and v9 though
            password = None
        settings.set_password(password)

    regex = re.compile('<input type="text" name="ssid" value="(.+?)"')
    match_object = regex.search(html_basic)
    if match_object is None:
        raise RouterParseError('Cannot determine SSID')
    settings.set_ssid(match_object.group(1))


    regex = re.compile('<input type="hidden" name="initChannel" value="([0-9]+)">')
    match_object = regex.search(html_basic)
    if match_object is None:
        raise RouterParseError('Cannot determine channel')
    settings.set_channel(int(match_object.group(1)))


    if '<input type="checkbox"  checked name="enable_ap" value="enable_ap">' in html_advanced:
        is_enabled = True
    else:
        is_enabled = False
    settings.set_enabled_status(is_enabled)


    if '<input type="checkbox"  checked name="ssid_bc" value="ssid_bc">' in html_advanced:
        is_broadcasting = True
    else:
        is_broadcasting = False
    settings.set_ssid_broadcast_status(is_broadcasting)


    if '<input type="checkbox"  checked name="enable_wmm" value="enable_wmm">' in html_advanced:
        is_wmm_enabled = True
    else:
        is_wmm_enabled = False
    settings.set_internal_param('enable_wmm', is_wmm_enabled)


    regex = re.compile("Select Region(?:.+?)<option selected value=\"([0-9]+)\">")
    match_object = regex.search(html_basic)
    if match_object is None:
        raise RouterParseError('Cannot determine Region value')
    settings.set_internal_param('WRegion', int(match_object.group(1)))

    return settings


def _parse_addr_reservation_list(html):
    lst = DHCPReservationList()
    lst.set_reboot_requirement_status(False)

    # Determine the submit token.. some WGR614v9 models have such a token
    # It's either some form of CSRF protection or something else..
    match_object = re.compile('<form name="frmLan" method="POST" action="lan.cgi\?id=([0-9]+)">').search(html)
    if match_object is None:
        lst.set_internal_param('submit_token', None)
    else:
        lst.set_internal_param('submit_token', int(match_object.group(1)))

    regex = '<tr>(?:.+?)<span class="ttext">(.+?)</span></td>(?:.+?)<span class="ttext">(?:.+?)</span></td>(?:.+?)<span class="ttext">(.+?)</span></td></tr>'
    for ip, mac in re.compile(regex).findall(html):
        item = DHCPReservationListItem()
        item.set_ip(ip)
        item.set_mac(mac)
        lst.append(item)

    return lst


def _parse_dhcp_settings(html):
    settings = DHCPServerSettings()
    settings.set_enabled_status('<INPUT name=lan_proto type=hidden value= "dhcp">' in html)

    match_object = re.compile('<INPUT name=dhcp_start type=hidden value= "(.+?)"').search(html)
    if match_object is None:
        raise RouterParseError('Cannot determine DHCP start IP')
    settings.set_ip_start(match_object.group(1))

    match_object = re.compile('<INPUT name=dhcp_end type=hidden value= "(.+?)"').search(html)
    if match_object is None:
        raise RouterParseError('Cannot determine DHCP end IP')
    settings.set_ip_end(match_object.group(1))

    settings.ensure_valid()

    return settings


def _generate_dmz_data_WGR614v9(settings):
    settings.ensure_valid()

    data = {}
    data['rspToPing'] = 'rspToPing'
    data['wan_mtu'] = 1480
    data['NatInboundFiltering'] = 'Secured'
    data['apply'] = 'Apply'
    data['wan_way'] = 1492
    data['nat_inbound_filtering'] = 1
    data['wan_proto'] = 'pppoe'
    data['nvram_mtu'] = 1480
    data['dmz_ip'] = settings.ip

    if settings.is_enabled:
        data['dmz_enable'] = 'dmz_enable'

    data['dmzip1'], data['dmzip2'], data['dmzip3'], data['dmzip4'] = map(int, settings.ip.split('.'))

    return data


def _generate_dmz_data_WGR614v8(settings):
    data = _generate_dmz_data_WGR614v9(settings)
    del data['dmz_ip']
    return data


def _generate_dmz_data_WGR614v7(settings):
    data = _generate_dmz_data_WGR614v8(settings)

    # Connect Automatically, as Required
    data['dod'] = 'dod'

    return data


def _generate_wireless_basic_settings_WGR614v9(settings):
    settings.ensure_valid()

    data = {}
    data['ssid'] = settings.ssid
    data['WRegion'] = settings.get_internal_param('WRegion')
    data['w_channel'] = settings.channel
    data['opmode'] = 'g and b'
    data['Apply'] = 'Apply'
    data['initChannel'] = 0 # some hidden field
    data['wds_enable'] = 0
    data['tempSetting'] = 0
    data['tempRegion'] = settings.get_internal_param('WRegion')

    if settings.security_type_is_wep:
        data['security_type'] = 'WEP'
        data['authAlgm'] = 'automatic'
        data['initAuthType'] = 'automatic'
        data['initDefaultKey'] = 0
        data['wepenc'] = 1 if settings.security_type == settings.__class__.SECURITY_TYPE_WEP64 else 2
        data['wep_key_no'] = 1 # the id of the WEP key to use (between 1 and 4)
        data['KEY1'] = settings.password
        data['KEY2'] = ''
        data['KEY3'] = ''
        data['KEY4'] = ''
        data['passphraseStr'] = '' # This is a temporary field used in the UI only
    elif settings.security_type_is_wpa:
        data['security_type'] = 'WPA-PSK' if settings.security_type == settings.__class__.SECURITY_TYPE_WPA else 'WPA2-PSK'
        data['passphrase'] = settings.password
        data['pfChanged'] = 1 # i don't know what this does
        data['tempSetting'] = 1 # i don't know what this does
    else:
        data['security_type'] = 'Disable'
        data['tempSetting'] = 1 # i don't know what this does

    return data


def _generate_wireless_advanced_settings_WGR614v9(settings):
    settings.ensure_valid()

    data = {}
    data['frag'] = 2346
    data['rts'] = 2347
    data['enable_shortpreamble'] = '' # empty means 'enable long preamble'
    data['Apply'] = 'Apply'
    data['enable_wmm'] = 'enable_wmm'

    if settings.is_enabled:
        data['enable_ap'] = 'enable_ap'

    if settings.is_broadcasting_ssid:
        data['ssid_bc'] = 'ssid_bc'

    return data


def _generate_wireless_advanced_settings_WGR614v8(settings):
    data = _generate_wireless_advanced_settings_WGR614v9(settings)
    data['wsc_config'] = 'on'
    data['wps_enable'] = 'enabled'
    return data


def _generate_wireless_advanced_settings_WGR614v7(settings):
    settings.ensure_valid()

    data = {}
    data['Apply'] = 'Apply'
    data['enable_wmm'] = 1

    if settings.is_enabled:
        data['enable_ap'] = 'enable_ap'

    if settings.is_broadcasting_ssid:
        data['ssid_bc'] = 'ssid_bc'

    return data


def _generate_addr_reservation_item_data(item):
    ip_parts = item.ip.split('.')

    data = {}
    data['apply'] = 'Add'
    data['dv_name'] = item.mac # device name, up to 20 chars
    data['rsv_ip'] = item.ip
    data['rsv_ip1'] = ip_parts[0]
    data['rsv_ip2'] = ip_parts[1]
    data['rsv_ip3'] = ip_parts[2]
    data['rsv_ip4'] = ip_parts[3]
    data['rsv_mac'] = item.mac.upper()
    data['rsv_mode'] = 'add'

    return data



from roscraco.helper import validator
from roscraco.exception import RouterSettingsError


class DMZSettings(object):
    """Represents the DMZ settings for a router."""

    #: List of properties to export using export()
    PROPERTIES = (
        'is_enabled', 'ip'
    )

    def __init__(self):
        self._is_supported = False
        self._is_enabled = False
        self._ip = None
        self._changes_require_reboot = True

    def set_reboot_requirement_status(self, value):
        self._changes_require_reboot = bool(value)

    @property
    def changes_require_reboot(self):
        """Tells whether the router needs rebooting
        for changes to take effect.
        """
        return self._changes_require_reboot

    @property
    def is_supported(self):
        """Tells whether the router supports DMZ."""
        return self._is_supported

    def set_supported_status(self, value):
        self._is_supported = bool(value)

    @property
    def is_enabled(self):
        """Tells whether the DMZ feature is enabled."""
        return self._is_enabled

    def set_enabled_status(self, value):
        self._is_enabled = bool(value)

    @property
    def ip(self):
        """The IP address of the DMZ host."""
        return self._ip

    def set_ip(self, value):
        self._ip = value

    def validate(self):
        errors = {}
        if not validator.is_valid_ip_address(self.ip):
            errors['ip'] = 'Invalid IP: %s' % self.ip
        return errors

    def ensure_valid(self):
        errors = self.validate()
        if len(errors) != 0:
            raise RouterSettingsError(str(errors))

    def __eq__(self, other):
        return self.ip == other.ip and self.is_enabled == other.is_enabled

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return id(self)

    def __repr__(self):
        mode = 'ENABLED' if self.is_enabled else 'DISABLED'
        return '<%s: %s/%s>' % (self.__class__, mode, self.ip)

    def export(self):
        """Exports the most important settings attributes,
        omitting any internal attributes.
        """
        export = {}
        for attr in self.__class__.PROPERTIES:
            export[attr] = getattr(self, attr, None)
        return export



import re

from roscraco.router.base import RouterBase

from roscraco.helper import validator, converter, split_in_groups
from roscraco.exception import RouterFetchError, RouterParseError, \
     RouterSettingsError, RouterIdentityError, RouterLoginError

from roscraco.response import RouterInfo, TrafficStats, DMZSettings, \
     ConnectedClientsListItem, ConnectedClientsList, WirelessSettings, \
     DHCPReservationList, DHCPReservationListItem, DHCPServerSettings


class Zyxel_P320W(RouterBase):

    def __init__(self, *args, **kwargs):
        RouterBase.__init__(self, *args, **kwargs)
        self._is_logged_in = False

    def _handle_first_request(self):
        """We need to make an initial "login" request.

        This usually happens via an HTML form.
        If we don't login and try to access anything other than `/`,
        we'll be redirected to http://192.168.1.1/

        Login works on an IP basis. There's no sessions (cookies) or anything.
        After we've logged in our session is active for some time.
        """
        data = {'rc': '@', 'rd': 'login', 'TC': '', 'PS': self.password}
        contents = self._make_http_request_write('cgi-bin/logi', data)
        if 'Error : Password is incorrect!' in contents:
            raise RouterLoginError('Cannot login - wrong password.')
        self._is_logged_in = True

    def close(self):
        # It's a good idea to perform a logout, because no one else
        # will be able to log in until the current session expires
        # (the session is on an IP basis)
        try:
            if self._is_logged_in:
                # Only logout if we've made at least one successful request
                self._make_http_request_read('cgi-bin/logo?rc=@&rd=login')
                self._is_logged_in = False
        except RouterFetchError:
            pass

    def confirm_identity(self):
        c = self._make_http_request_read('')
        # Both of these strings are valid P-320W routers
        # <TITLE>.:: Welcome to ZyXEL P-320W (P-320W) ::. </TITLE>
        # <TITLE>.:: Welcome to ZyXEL P-320W (router) ::. </TITLE>
        if '<TITLE>.:: Welcome to ZyXEL P-320W' not in c:
            raise RouterIdentityError('Cannot confirm identity')

    def get_router_info(self):
        return parse_router_info(self._make_http_request_read('fwup.htm'))

    def get_uptime(self):
        return parse_uptime(self._make_http_request_read('status.htm'))

    def get_pppoe_online_time(self):
        return None

    def get_traffic_stats(self):
        return parse_traffic_stats(self._make_http_request_read('slist.htm'))

    def get_mac_address(self):
        return parse_mac_address(self._make_http_request_read('prim.htm'))

    def get_dns_servers(self):
        return parse_dns_servers(self._make_http_request_read('status.htm'))

    def get_connected_clients_list(self):
        return parse_connected_clients_list(self._make_http_request_read('clist.htm'))

    def get_dmz_settings(self):
        settings = DMZSettings()
        settings.set_supported_status(False)
        settings.set_enabled_status(False)
        settings.set_ip('0.0.0.0')
        return settings

    def push_dmz_settings(self, settings):
        return False

    def get_dhcp_settings(self):
        return parse_dhcp_settings(self._make_http_request_read('dhcp.htm'))

    def get_addr_reservation_list(self):
        return parse_addr_reservation_list(self._make_http_request_read('sdhcp.htm'))

    def push_addr_reservation_list(self, lst_new):
        lst_new.ensure_valid()

        lst_old = self.get_addr_reservation_list()
        if lst_old == lst_new:
            return True

        data = generate_addr_reservation_data(lst_new)
        self._make_http_request_write('cgi-bin/dhcp', data)
        return self.get_addr_reservation_list() == lst_new

    def get_wireless_settings(self):
        main = self._make_http_request_read('main.htm')
        security_type, link = generate_wireless_settings_link(main)
        html = self._make_http_request_read(link)
        return parse_wireless_setting(security_type, html)

    def push_wireless_settings(self, settings):
        data = generate_wireless_data(settings)

        # When updating WPA settings the router tries to generate some KEY
        # or whatever, which slows things down (~15 seconds).
        # For non-WPA update requests, we leave the default timeout
        timeout = 20.0 if settings.security_type_is_wpa else None
        self._make_http_request_write('cgi-bin/wlap', data, timeout=timeout)
        return self.get_wireless_settings() == settings

    @property
    def supports_reboot(self):
        return True

    def reboot(self):
        url = 'cgi-bin/rebo?D=&rd=space'
        self._make_http_request_write(url)


def parse_router_info(html):
    obj = RouterInfo()
    obj.set_hardware_version('P320W')

    regex_firmware = re.compile('fwug\[4\]\+" (.+?) "\+fwug\[5\]', re.DOTALL)
    match_object = regex_firmware.search(html)
    if match_object is None:
        raise RouterParseError('Cannot parse firmware version')
    obj.set_firmware_version(match_object.group(1))

    return obj


def parse_uptime(html):
    regex = re.compile('sta\[22\]\+" : ","([0-9]+):([0-9]+):([0-9]+)",')
    match_object = regex.search(html)
    if match_object is None:
        raise RouterParseError('Cannot parse uptime')
    hours, minutes, seconds = map(int, match_object.groups())
    return hours * 3600 + minutes * 60 + seconds


def parse_traffic_stats(html):
    regex = re.compile('slist\[0\]\+"</B>","([0-9]+)","([0-9]+)",')
    match_object = regex.search(html)
    if match_object is None:
        raise RouterParseError('Cannot parse traffic stats')
    packets_recv, packets_sent = map(int, match_object.groups())
    return TrafficStats(0, 0, packets_recv, packets_sent)


def parse_mac_address(html):
    regex = re.compile('NAME=_En VALUE="(.+?)"')
    match_object = regex.search(html)
    if match_object is None:
        raise RouterParseError('Cannot parse mac address')
    return converter.normalize_mac(match_object.group(1))


def parse_dns_servers(html):
    regex = re.compile('sta\[11\]\+" : ","(.+?)"')
    match_object = regex.search(html)
    if match_object is None:
        raise RouterParseError('Cannot parse dns servers')
    dns_ips = match_object.group(1).split(', ')
    return [ip for ip in dns_ips if validator.is_valid_ip_address(ip)]


def parse_connected_clients_list(html):
    lst = ConnectedClientsList()

    regex = re.compile('var x=new Array\((.+?)\);')
    match_object = regex.search(html)
    if match_object is None:
        raise RouterParseError('Cannot parse connected clients list')
    match_str = match_object.group(1)
    if match_str == '""':
        return lst

    import ast
    try:
        list_literal = '[%s]' % match_str
        list_items = ast.literal_eval(list_literal)
        list_items.pop() # empty 'sentinel' string - not needed
        list_items = [v.decode('utf-8', 'ignore') if isinstance(v, bytes) else v
                      for v in list_items]
        for ip, name, mac in split_in_groups(list_items, 3):
            item = ConnectedClientsListItem()
            item.set_client_name(name)
            item.set_mac(converter.normalize_mac(mac))
            item.set_ip(ip)
            item.set_lease_time(item.__class__.LEASE_TIME_PERMANENT)
            lst.append(item)
        return lst
    except Exception:
        return lst


def parse_dhcp_settings(html):
    is_enabled = '<INPUT TYPE=CHECKBOX NAME=_HE CHECKED>' in html

    # 192.168.1 is hardcoded.. only the last part changes
    regex = re.compile('<INPUT NAME=HR0 VALUE="([0-9]+)" SIZE=3 MAXLENGTH=3>')
    match_object = regex.search(html)
    if match_object is None:
        raise RouterParseError('Cannot parse dhcp start ip!')
    ip_last_part = match_object.group(1)
    ip_start = '192.168.1.%s' % ip_last_part

    # get the pool size to determine what the last IP is
    regex = re.compile('document.forms\[0\]\._HR1\.value=([0-9]+)-([0-9]+)')
    match_object = regex.search(html)
    if match_object is None:
        raise RouterParseError('Cannot parse dhcp pool size!')
    num1, num2 = map(int, match_object.groups())
    pool_size = num1 - num2
    ip_start_long = converter.ip2long(ip_start)
    ip_end_long = ip_start_long + pool_size
    ip_end = converter.long2ip(ip_end_long)

    settings = DHCPServerSettings()
    settings.set_enabled_status(is_enabled)
    settings.set_ip_start(ip_start)
    settings.set_ip_end(ip_end)
    settings.ensure_valid()
    return settings


def parse_addr_reservation_list(html):
    """
    Entries look like this:
        {status?}, {mac}, {ip_last_part}, {optional description}
        "1","00-11-11-11-11-11","52","",
        "1","00-11-11-11-11-11","52","R_user-cc5c756950",

    The status in front (0 or 1) is not important.
    It's 0 for all entries, when the list is empty,
    and 1 when there's at least one real entry.

    The IP prefix is always 192.168.1.

    The optional description only appears in entries
    not added manually, but automatically via a 'Reserve' checkbox in
    the connected clients list
    """
    regex = re.compile('^"0|1","(.+?)","([0-9]+)","(?:.*?)",\n')
    reservation_list = DHCPReservationList()
    reservation_list.set_reboot_requirement_status(False)
    for mac, ip_last_part in regex.findall(html):
        if ip_last_part == '0':
            continue
        ip = '192.168.1.%s' % ip_last_part
        item = DHCPReservationListItem()
        item.set_mac(converter.normalize_mac(mac))
        item.set_ip(ip)
        item.set_enabled_status(True)
        reservation_list.append(item)
    return reservation_list


def _resolve_wireless_security_type(cskm):
    """The ``cskm`` variable seen on every page,
    specifies the type of wireless security in use now.

    Depending on that variable, a unique wireless settings
    link is generated.
    """

    # cskm_value => (web link value, security_type)
    cskm_map = {}

    cskm_map['0'] = ('0000', WirelessSettings.SECURITY_TYPE_NONE)

    # WEP could either be WEP64 or WEP128..
    # there's no way we can determine that from this context..
    cskm_map['1'] = ('1000', 'wep')

    #Dynamic WEP (radius) - not supported
    cskm_map['1CHECKED'] = ('2000', WirelessSettings.SECURITY_TYPE_NONE)

    # WPA radius - not supported
    cskm_map['2CHECKED'] = ('8000', WirelessSettings.SECURITY_TYPE_NONE)
    cskm_map['3CHECKED'] = ('8000', WirelessSettings.SECURITY_TYPE_NONE)

    # WPA-PSK
    cskm_map['2'] = ('4000', WirelessSettings.SECURITY_TYPE_WPA)
    cskm_map['3'] = ('4000', WirelessSettings.SECURITY_TYPE_WPA)

    # No security, but there's also a Pre-Shared key on the page
    # Anyways, this is not supported
    cskm_map['4'] = ('0100', WirelessSettings.SECURITY_TYPE_NONE)
    cskm_map['5'] = ('0100', WirelessSettings.SECURITY_TYPE_NONE)

    # No security, but there's some radius settings on the page
    # Anyways, this is not supported
    cskm_map['4CHECKED'] = ('0200', WirelessSettings.SECURITY_TYPE_NONE)
    cskm_map['5CHECKED'] = ('0200', WirelessSettings.SECURITY_TYPE_NONE)

    return cskm_map.get(cskm, cskm_map.get('0'))


def generate_wireless_settings_link(html):
    """The link with the wireless settings is generated
    dynamically depending on 2 variables find on every page.
    These are the ``es0`` and ``cskm`` variables.
    What ``es0`` controls is not yet known, but ``cskm``
    represents the type of wireless security that's setup now.
    """
    regex = re.compile('es0="([0-9]+)";(?:.+?)cskm="(.+?)";', re.DOTALL)
    match_object = regex.search(html)
    if match_object is None:
        raise RouterParseError('Cannot parse wireless link variables!')
    es0, cskm = match_object.groups()
    link_value, security_type = _resolve_wireless_security_type(cskm)
    from time import time
    return security_type, 'wlan.htm?rc=&rf=%s&_=&_=&_=&_=&_=%s&_=&ZT=%d' % (
        link_value,
        es0,
        time()
    )


def parse_wireless_setting(security_type, html):
    obj = WirelessSettings()
    obj.add_security_support(WirelessSettings.SECURITY_TYPE_WEP64)
    obj.add_security_support(WirelessSettings.SECURITY_TYPE_WEP128)
    obj.add_security_support(WirelessSettings.SECURITY_TYPE_WPA)
    obj.set_auto_channel_support(False)
    obj.set_reboot_requirement_status(False)

    regex = re.compile('if \(""==""\) document\.forms\[0\]\.ZN\.value='
                       'unescape\("(.+?)"\);')
    match_object = regex.search(html)
    if match_object is None:
        raise RouterParseError('Cannot parse ssid!')
    ssid = match_object.group(1)
    obj.set_ssid(ssid)

    # security_type is either a valid value or 'wep',
    # in which case we've got to resolve it to WEP64 or WEP128
    if security_type == 'wep':
        regex = re.compile('(?:.+?)<OPTION CHECKED>(?:.+?)dw\(wlan\[25\]\)'
                           '(?:.+?)</SCRIPT>', re.DOTALL)
        if regex.search(html) is not None:
            security_type = WirelessSettings.SECURITY_TYPE_WEP64
        else:
            security_type = WirelessSettings.SECURITY_TYPE_WEP128

        regex = re.compile('<INPUT NAME=ZO0 VALUE="(.+?)" SIZE=26')
        match_object = regex.search(html)
        if match_object is None:
            raise RouterParseError('Cannot parse WEP password!')
        obj.set_password(match_object.group(1))
    elif security_type == WirelessSettings.SECURITY_TYPE_WPA:
        regex = re.compile('document\.forms\[0\]\.PK\.value=unescape\("(.+?)"\);')
        match_object = regex.search(html)
        if match_object is None:
            raise RouterParseError('Cannot parse WPA password!')
        obj.set_password(match_object.group(1))
    else: # No security
        obj.set_password('')
    obj.set_security_type(security_type)

    regex = re.compile('if\(i==([0-9]+)\)\nsel="SELECTED"')
    match_object = regex.search(html)
    if match_object is None:
        raise RouterParseError('Cannot parse channel!')
    obj.set_channel(int(match_object.group(1)))

    # `_esv` is either 0 or 1 and controls SSID broadcasting
    if '_esv=1' in html:
        obj.set_ssid_broadcast_status(False)

    regex = re.compile('wdv0=\("(.+?)"=="(.+?)"\)\?true:false')
    match_object = regex.search(html)
    if match_object is None:
        raise RouterParseError('Cannot parse enabled status!')
    val1, val2 = match_object.groups()
    obj.set_enabled_status(val1 == val2)

    return obj


def _denormalize_mac(mac):
    """Takes a normalized mac address (all lowercase hex, no separators)
    and converts it to the Zyxel format.

    Example::
        _denormalize_mac('abcdef123456') == 'ab-cd-ef-12-34-56'
    """
    return '-'.join((mac[i] + mac[i+1] for i in range(0, 12, 2)))


def generate_addr_reservation_data(lst):
    """The POST data we need to send contains the
    information about all reservation list items.
    """
    slots = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H']
    if len(slots) < len(lst):
        raise RouterSettingsError('More entries than slots!')

    data = {}
    data['Eb'] = 0
    data['rc'] = '@'
    data['rd'] = 'sdhcp'

    # Set some default data for all fields
    for slot_id in slots:
        data['DE' + slot_id] = 'o'
        data['DN' + slot_id] = ''
        data['Di' + slot_id] = 0 # last part of IP
        data['Dm' + slot_id] = '00-00-00-00-00-00' # MAC address

    for i, item in enumerate(lst):
        slot_id = slots[i]
        data['Di' + slot_id] = item.ip.split('.').pop()
        data['Dm' + slot_id] = _denormalize_mac(item.mac)

    return data


def generate_wireless_data(settings):
    """This is very ugly due to the fact that this incredibly
    stupid router expects the fields to come in order.

    If we change the order of them, the settings get corrupted,
    or rejected.

    This means that a dictionary cannot be used here,
    because its order is not guaranteed.
    Some fields (such as KM5) also need to be sent twice,
    which can't be done with a dictionary.
    """
    settings.ensure_valid()
    if settings.security_type_is_wep:
        CS = '1'
        rf = '1000'
    elif settings.security_type_is_wpa:
        CS = '2'
        rf = '4000'
    else:
        CS = '0'
        rf = '0000'

    if not settings.is_enabled:
        rf = ''

    data = []
    data.append(('RC', '@'))
    data.append(('rf', rf))
    data.append(('Xf', '1'))
    data.append(('prev', ''))

    data.append(('_ZN', ''))
    data.append(('_ZC', ''))
    data.append(('_ESS', ''))
    data.append(('_WD', ''))
    data.append(('CS', CS))

    if settings.is_enabled:
        data.append(('WD', 'o'))
        data.append(('xWD', 'on'))
    else:
        data.append(('WD', 'x'))

    data.append(('ZN', settings.ssid))

    data.append(('es', 'FFFC'))

    if not settings.is_broadcasting_ssid:
        data.append(('ES0', 'o'))
        data.append(('ES1', 'o'))
        data.append(('_ES', 'on'))

    data.append(('ZC', settings.channel))
    data.append(('_Security', ''))

    if settings.security_type_is_wep:
        data.append(('KM#3', '000F'))
        data.append(('KM0', 'x'))
        data.append(('ZP', ''))
        if settings.security_type == settings.__class__.SECURITY_TYPE_WEP64:
            zs = 0
            _ZS = '64-bit WEP'
        else:
            zs = 1
            _ZS = '128-bit WEP'
        if settings.is_wep_password_in_hex:
            ZS = zs + 1
        else:
            ZS = 128 + zs + 1
        data.append(('ZS', ZS))
        data.append(('_ZS', _ZS))
        # Authentication method = Both
        data.append(('AU', '2'))
        # related to ASCII/HEX radio buttons
        data.append(('_MZS', 'on'))
        # which WEP key to use
        data.append(('ZW', '0'))
        data.append(('ZO0', settings.password))
        data.append(('ZO1', ''))
        data.append(('ZO2', ''))
        data.append(('ZO3', ''))
    elif settings.security_type_is_wpa:
        data.append(('KM#3', '00FF'))
        data.append(('KM4', 'o'))
        data.append(('KM5', ''))
        data.append(('PK', settings.password))

    data.append(('KM5', ''))
    data.append(('rd', 'wlan'))

    return data



import urllib

from roscraco.response import WirelessSettings
from base import TplinkBase, _extract_js_array_data


class Tplink_WR740N(TplinkBase):

    def confirm_identity(self):
        self._ensure_www_auth_header('Basic realm="TP-LINK Wireless Lite N Router WR740N"')

    def get_wireless_settings(self):
        html_settings = self._make_http_request_read('WlanNetworkRpm.htm')
        html_security = self._make_http_request_read('WlanSecurityRpm.htm')
        return _parse_wireless_settings(html_settings, html_security)

    def push_wireless_settings(self, settings):
        get_params_settings = _generate_wireless_settings_data(settings)
        self._make_http_request_write('WlanNetworkRpm.htm?%s' % urllib.urlencode(get_params_settings))

        get_params_security = _generate_wireless_security_data(settings)
        self._make_http_request_write('WlanSecurityRpm.htm?%s' % urllib.urlencode(get_params_security))

        return self.get_wireless_settings() == settings


def _parse_wireless_settings(html_settings, html_security):
    """Extracts the Wireless settings from the page contents
    for a WR740N (and some other models)."""
    obj = WirelessSettings()
    obj.add_security_support(WirelessSettings.SECURITY_TYPE_WEP64)
    obj.add_security_support(WirelessSettings.SECURITY_TYPE_WEP128)
    obj.add_security_support(WirelessSettings.SECURITY_TYPE_WPA)
    obj.add_security_support(WirelessSettings.SECURITY_TYPE_WPA2)

    settings_array = _extract_js_array_data(html_settings, 'wlanPara')
    obj.set_ssid(settings_array[3])
    obj.set_enabled_status(settings_array[8] == 1)
    obj.set_ssid_broadcast_status(settings_array[9] == 1)
    obj.set_channel(settings_array[10])
    obj.set_internal_param('region', settings_array[5])
    obj.set_internal_param('mode', settings_array[7])
    obj.set_internal_param('chanWidth', settings_array[11])
    obj.set_internal_param('rate', settings_array[12])

    security_array = _extract_js_array_data(html_security, 'wlanPara')
    wlan_list_array = _extract_js_array_data(html_security, 'wlanList')

    security_type = int(security_array[2])

    if security_type == 1: # WEP of some sort
        bit_length = int(wlan_list_array[1])
        if bit_length == 13: # 128 bit length
            security_type = WirelessSettings.SECURITY_TYPE_WEP128
        else:
            security_type = WirelessSettings.SECURITY_TYPE_WEP64
    elif security_type == 3: # WPA-PSK (WPA or WPA2)
        # string like '331', '332', '333'
        # we're interested in the 3rd char, which deals with WPA-PSK/WPA2-PSK
        # 3rd char possible values {1: WPA, 2: WPA2, 3: Automatic (WPA + WPA2)}
        security_options = security_array[3]

        if int(security_options[2]) == 1:
            security_type = WirelessSettings.SECURITY_TYPE_WPA
        else:
            security_type = WirelessSettings.SECURITY_TYPE_WPA2
    else: # type is either 0 (no security) or 2 (WPA-Enterprise)
        security_type = WirelessSettings.SECURITY_TYPE_NONE

    obj.set_security_type(security_type)

    password = wlan_list_array[0] if obj.security_type_is_wep else security_array[9]
    obj.set_password(password)

    return obj


def _generate_wireless_security_data(settings):
    """Generates a wireless security data array to push to the
    router from the given wireless settings.

    Keep in mind that the settings array generated by this is resetting some fields
    that it doesn't support/care about (like radius settings, etc.).
    """
    WirelessSettings = settings.__class__

    settings.ensure_valid()

    data = {}
    data['wepSecOpt'] = 3 # WEP Type = Automatic
    data['wpaSecOpt'] = 3 # WPA-Enterprise Type = Automatic
    data['pskSecOpt'] = 1 if settings.security_type == WirelessSettings.SECURITY_TYPE_WPA else 2 # 1 = WPA, 2 = WPA2
    data['pskCipher'] = 1 # WPA-PSK/WPA2-PSK encryption method (1: Automatic, 2: TKIP, 3: AES)
    data['wpaCipher'] = 1 # WPA-Enterprise encryption method (1: Automatic)
    data['pskSecret'] = '' # default value for a WPA-PSK/WPA2-PSK password
    data['keytype'] = 2 # 1 is for HEX WEP, 2 is for ASCII WEP password
    data['keynum'] = 1 # which of the 4 WEP keys to use (if WEP is enabled)

    # These are the 4 WEP keys and their lengths.. empty by default - to be filled later if needed
    # only one of these keys is actively used (specified by `keynum` above)
    merge_with =  {'key1': '', 'key2': '', 'key3': '', 'key4': '', 'length1': 0, 'length2': 0, 'length3': 0, 'length4': 0}
    data = dict(data, **merge_with)

    # Some more WPA-Enterprise settings (we're not using those at all)
    merge_with = {'radiusIp': '', 'radiusPort': 1812, 'radiusSecret': ''}
    data = dict(data, **merge_with)

    # WPA and WPA-PSK intervals and Save button field
    merge_with = {'intervalWpa': 0, 'interval': 0, 'Save': 'Save'}
    data = dict(data, **merge_with)


    if settings.security_type_is_wep:
        data['secType'] = 1
    elif settings.security_type in (WirelessSettings.SECURITY_TYPE_WPA, WirelessSettings.SECURITY_TYPE_WPA2):
        data['secType'] = 3
    else:
        data['secType'] = 0


    if settings.security_type_is_wep:
        # 64bit = 5, 128bit = 13
        bit_length = 5 if settings.security_type == WirelessSettings.SECURITY_TYPE_WEP64 else 13

        data['key1'] = settings.password
        data['length1'] = bit_length
        data['keytype'] = 1 if settings.is_wep_password_in_hex else 2 # ASCII or HEX
    elif settings.security_type != WirelessSettings.SECURITY_TYPE_NONE:
        data['pskSecret'] = settings.password


    return data


def _generate_wireless_settings_data(settings):
    """Generates a wireless settings data array to push to the
    router from the given wireless settings."""
    settings.ensure_valid()

    data = {'ssid1': settings.ssid, 'channel': settings.channel, 'Save': 'Save'}

    if settings.is_enabled:
        data['ap'] = 1

    if settings.is_broadcasting_ssid:
        data['broadcast'] = 2

    # preserve some of the params we don't handle
    for k in ('region', 'mode', 'chanWidth', 'rate'):
        data[k] = settings.get_internal_param(k)

    # WDS (WLAN bridging) related settings
    # we'll clear them all by default
    merge_with = {'brlssid': '', 'brlbssid': '', 'keytype': 1, 'wepindex'
                    'authtype': 1, 'keytext': ''}
    data = dict(data, **merge_with)

    return data



import re
import base64

from roscraco.router.base import RouterBase

from roscraco.helper import validator, converter
from roscraco.exception import RouterParseError, RouterIdentityError

from roscraco.response import RouterInfo, TrafficStats, DMZSettings, \
     ConnectedClientsListItem, ConnectedClientsList, WirelessSettings, \
     DHCPReservationList, DHCPServerSettings


class Canyon_CNWF514(RouterBase):

    def _perform_http_request(self, *args, **kwargs):
        auth = base64.b64encode('%s:%s' % (self.username, self.password))
        kwargs['headers'] = [('Authorization', 'Basic %s' % auth)]
        return RouterBase._perform_http_request(self, *args, **kwargs)

    def confirm_identity(self):
        _, headers, contents = self._perform_http_request('%shome.asp' %
                                                          self.url_base)
        header_server = headers.getheader('Server')
        if header_server != 'GoAhead-Webs':
            raise RouterIdentityError('Invalid Server header: %s' % header_server)

        string_to_find = '<! Copyright (c) Realtek Semiconductor Corp., 2003. All Rights Reserved. ->'
        if string_to_find not in contents:
            raise RouterIdentityError('Cannot find string in contents: %s' % string_to_find)

    def get_router_info(self):
        return _parse_router_info(self._make_http_request_read('status.asp'))

    def get_pppoe_online_time(self):
        return None  # online time not specified anywhere in the interface

    def get_uptime(self):
        return _parse_uptime_to_seconds(self._make_http_request_read('status.asp'))

    def get_traffic_stats(self):
        return _parse_traffic_stats(self._make_http_request_read('stats.asp'))

    def get_mac_address(self):
        return _parse_mac_address(self._make_http_request_read('status.asp'))

    def get_dns_servers(self):
        return _parse_dns_servers(self._make_http_request_read('tcpipwan.asp'))

    def get_connected_clients_list(self):
        return _parse_connected_clients_list(self._make_http_request_read('dhcptbl.asp'))

    def get_wireless_settings(self):
        html = [self._make_http_request_read(path) for path in ('wlbasic.asp', 'wladvanced.asp', 'wlwpa.asp', 'wlwep.asp')]
        return _parse_wireless_settings(*html)

    def push_wireless_settings(self, settings):
        data_basic = _generate_wireless_data_basic(settings)
        self._make_http_request_write('goform/formWlanSetup', data_basic)

        data_advanced = _generate_wireless_data_advanced(settings)
        self._make_http_request_write('goform/formAdvanceSetup', data_advanced)

        if settings.security_type_is_wep:
            data_wep = _generate_wireless_data_wep(settings)
            self._make_http_request_write('goform/formWep', data_wep)

        data_security = _generate_wireless_data_security(settings)
        self._make_http_request_write('goform/formWlEncrypt', data_security)

        return settings.eq(self.get_wireless_settings(), skip_attrs=('password',))

    def get_dmz_settings(self):
        # DMZ breaks remote management for this router..
        # so we're not enabling support for it at all
        settings = DMZSettings()
        settings.set_supported_status(False)
        settings.set_enabled_status(False)
        settings.set_ip('0.0.0.0')
        return settings

    def push_dmz_settings(self, settings):
        # DMZ is broken for this router,
        # refusing to do anything
        return False

    def get_addr_reservation_list(self):
        lst = DHCPReservationList()
        lst.set_reservation_support_status(False)
        lst.set_reboot_requirement_status(False)
        return lst

    def push_addr_reservation_list(self, lst):
        return False

    def get_dhcp_settings(self):
        return _parse_dhcp_settings(self._make_http_request_read('tcpiplan.asp'))

    @property
    def supports_reboot(self):
        return False

    def reboot(self):
        pass


def _parse_router_info(contents):
    obj = RouterInfo()
    obj.set_firmware_version(_parse_firmware_version(contents))
    obj.set_hardware_version("Unknown")
    return obj


def _parse_firmware_version(html):
    regex = "Firmware Version</b></td>(?:.+?)<td width=60%><font size=2>([a-zA-Z0-9-\.\(\)\s,]+)</td>"
    regex_firmware = re.compile(regex, re.DOTALL)

    match_object = regex_firmware.search(html)
    if match_object is None:
        raise RouterParseError("Cannot _parse firmware version")

    return match_object.group(1)


def _parse_uptime_to_seconds(html):
    regex = "Uptime</b></td>(?:.+?)<td width=60%><font size=2>([0-9]+)day:([0-9]+)h:([0-9]+)m:([0-9]+)s</td>"
    regex_uptime = re.compile(regex, re.DOTALL)

    match_object = regex_uptime.search(html)

    if match_object is None:
        raise RouterParseError("Cannot _parse uptime")

    days, hours, minutes, seconds = map(int, match_object.groups())

    return days * 86400 + hours * 3600 + minutes * 60 + seconds


def _parse_traffic_stats(contents):
    regex = "Ethernet WAN(?:.+?)Sent Packets(?:.+?)<font size=2>([0-9]+)</td>(?:.+?)Received Packets(?:.+?)<font size=2>([0-9]+)</td>"
    regex_traffic_stats = re.compile(regex, re.DOTALL)

    match_object = regex_traffic_stats.search(contents)
    if match_object is None:
        raise RouterParseError("Cannot _parse traffic stats")

    packets_recv, packets_sent = map(int, match_object.groups())
    # sadly, we've got only packets and no size information..
    return TrafficStats(0, 0, packets_recv, packets_sent)


def _parse_mac_address(contents):
    regex = "bssid_drv\[0\] ='((?:(?:[a-f0-9]{2}:){5})(?:[a-f0-9]{2}))';"
    regex_mac_address = re.compile(regex, re.DOTALL)

    match_object = regex_mac_address.search(contents)

    if match_object is None:
        raise RouterParseError("Cannot _parse mac address from contents")

    mac = match_object.group(1).strip(" ")
    if not validator.is_valid_mac_address(mac):
        raise RouterParseError("Found an invalid MAC address: %s" % mac)

    return converter.normalize_mac(mac)


def _parse_dns_servers(contents):
    regex = "name=\"dns[0-9]\"(?:.+?)value=(.*?)></td>"
    regex_dns_servers = re.compile(regex, re.DOTALL)

    servers = re.findall(regex_dns_servers, contents)
    return [ip.strip(" ") for ip in servers if validator.is_valid_ip_address(ip)]


def _parse_connected_clients_list(contents):
    regex = "<form(?:.+?)<tr(?:.+?)<tr bgcolor=(?:.+?)<font size=2>(.+?)</td><td><font size=2>(.+?)</td><td><font size=2>([0-9]+)</td>"
    regex_dhcp_list = re.compile(regex, re.DOTALL)

    lst = ConnectedClientsList()

    for id, match_groups in enumerate(re.findall(regex_dhcp_list, contents), start=1):
        ip, mac, lease_time = match_groups

        if not validator.is_valid_ip_address(ip):
            raise RouterParseError("Invalid IP address: %s" % ip)

        if not validator.is_valid_mac_address(mac):
            raise RouterParseError("Invalid MAC address: %s" % mac)

        lease_time = int(lease_time)

        item = ConnectedClientsListItem()
        item.set_client_name("Client %d" % id)
        item.set_mac(converter.normalize_mac(mac))
        item.set_ip(ip)

        item.set_lease_time(lease_time)

        lst.append(item)

    return lst


def _parse_wireless_settings(html_basic, html_advanced, html_security, html_wep):
    settings = WirelessSettings()
    settings.add_security_support(WirelessSettings.SECURITY_TYPE_WEP64)
    settings.add_security_support(WirelessSettings.SECURITY_TYPE_WEP128)
    settings.add_security_support(WirelessSettings.SECURITY_TYPE_WPA)
    settings.add_security_support(WirelessSettings.SECURITY_TYPE_WPA2)


    match_object = re.compile("var wps_ssid_old='(.+?)';").search(html_basic)
    if match_object is None:
        raise RouterParseError("Cannot find SSID!")
    settings.set_ssid(match_object.group(1))

    match_object = re.compile("var wps_disabled=(0|1);").search(html_basic)
    if match_object is None:
        raise RouterParseError("Cannot determine wireless enabled status!")
    settings.set_enabled_status(match_object.group(1) == "0")

    match_object = re.compile("defaultChan\[wlan_idx\]=(.+?);").search(html_basic)
    if match_object is None:
        raise RouterParseError("Cannot determine wireless channel!")
    settings.set_channel(int(match_object.group(1)))

    if "name=\"hiddenSSID\" value=\"no\"checked" not in html_advanced:
        settings.set_ssid_broadcast_status(False)


    # This is the security type (WEP, WPA, WPA2..)
    match_object = re.compile("var wps_encrypt_old=([0-4]);").search(html_security)
    if match_object is None:
        raise RouterParseError("Cannot determine security type!")
    security_type = int(match_object.group(1))

    if security_type == 1: # WEP
        match_object = re.compile("var wps_wep_keylen_old='([1-2])';").search(html_wep)
        if match_object is None:
            raise RouterParseError("Cannot determine WEP key length!")

        if int(match_object.group(1)) == 1: # 64bit
            settings.set_security_type(settings.__class__.SECURITY_TYPE_WEP64)
        else: # 128bit or something new that we don't handle
            settings.set_security_type(settings.__class__.SECURITY_TYPE_WEP128)
    elif security_type == 2: # WPA-PSK
        settings.set_security_type(settings.__class__.SECURITY_TYPE_WPA)
    elif security_type == 4: # WPA2-PSK
        settings.set_security_type(settings.__class__.SECURITY_TYPE_WPA2)
    else: # Either 0=No security or something else, which we don't handle
        settings.set_security_type(settings.__class__.SECURITY_TYPE_NONE)

    if settings.security_type_is_wpa:
        match_object = re.compile("var wps_psk_old='(.+?)';").search(html_security)
        if match_object is None:
            raise RouterParseError('Cannot determine wireless password!')
        settings.set_password(match_object.group(1))
    elif settings.security_type_is_wep:
        # WEP passwords are rendered as '****' on the page
        settings.set_password(None)
    else: # No security or something else
        settings.set_password("")

    return settings


def _parse_dhcp_settings(html):
    settings = DHCPServerSettings()
    settings.set_enabled_status('<option selected value="2">Server</option>' in html)

    match_object = re.compile('<input type="text" name="dhcpRangeStart" size="12" maxlength="15" value="(.+?)">').search(html)
    if match_object is None:
        raise RouterParseError("Cannot determine DHCP start IP")
    settings.set_ip_start(match_object.group(1))

    match_object = re.compile('<input type="text" name="dhcpRangeEnd" size="12" maxlength="15" value="(.+?)">').search(html)
    if match_object is None:
        raise RouterParseError("Cannot determine DHCP end IP")
    settings.set_ip_end(match_object.group(1))

    settings.ensure_valid()

    return settings


def _generate_wireless_data_basic(settings):
    settings.ensure_valid()

    data = {}

    data['wps_clear_configure_by_reg0'] = 0 # this is a hardcoded hidden field.. i don't know what it's for
    data['wlan-url'] = '/wlbasic.asp' # this is the 'redirect here when finished' url
    data['save'] = 'Save Settings'
    data['basicrates0'] = 0 # this is a hardcoded hidden field.. i don't know what it's for
    data['operrates0'] = 0 # this is a hardcoded hidden field.. i don't know what it's for

    data['band0'] = 2 # 2 means B+G mode
    data['mode0'] = 0 # 0 means 'operate in AccessPoint mode'
    data['ssid0'] = settings.ssid
    data['chan0'] = settings.channel

    if not settings.is_enabled:
        data['wlanDisabled0'] = 'ON'

    return data


def _generate_wireless_data_advanced(settings):
    settings.ensure_valid()

    data = {}

    data['authType'] = 'both' # Authentication Type (could be open/shared/both)
    data['fragThreshold'] = 2346 # Fragment Threshold (int from 256 to 2346)
    data['rtsThreshold'] = 2347 # RTS Threshold (int from 0 to 2347)
    data['beaconInterval'] = 100 # Beacon interval in milliseconds
    data['txRate'] = 0 # Data rate speed (0 = Auto)
    data['preamble'] = 'long' # Preamble Type (could be long/short)
    data['hiddenSSID'] = 'no' if settings.is_broadcasting_ssid else 'yes'
    data['iapp'] = 'yes' # IAPP (could be yes/no)
    data['11g_protection'] = 'yes' # 802.11g Protection (could be yes/no)
    data['wmm'] = 'off' # WMM (could be on/off)
    data['RFPower'] = 0 # 0 means 100% power
    data['turbo'] = 'auto' # Turbo Mode (could auto/always/off)
    data['save'] = 'Save Settings'
    data['submit-url'] = '/wladvanced.asp'

    return data


def _generate_wireless_data_security(settings):
    settings.ensure_valid()

    data = {}
    data['wps_clear_configure_by_reg0'] = '0'
    data['submit-url'] = '/wlwpa.asp'
    data['save'] = 'Save Settings'

    if settings.security_type_is_wep:
        data['method0'] = 1
    elif settings.security_type_is_wpa:
        # specifies whether we're using Personal or Radius auth
        data['wpaAuth0'] = 'psk'
        data['pskFormat0'] = '0' # 0 = passphrase; 1 = hex passphrase only
        data['pskValue0'] = settings.password
        if settings.security_type == settings.__class__.SECURITY_TYPE_WPA:
            data['method0'] = 2
            data['ciphersuite0'] = 'tkip'
        else:
            data['method0'] = 4
            data['wpa2ciphersuite0'] = 'aes'
    else: # no security
        data['method0'] = 0

    return data


def _generate_wireless_data_wep(settings):
    settings.ensure_valid()

    if not settings.security_type_is_wep:
        raise RuntimeError("Cannot generate WEP data from non-WEP settings")

    data = {}
    data['wps_clear_configure_by_reg0'] = '0'
    data['submit-url'] = '/wlwep.asp'
    data['save'] = 'Save Settings'
    # I think this field has the side effect
    # of enabling WEP, if it was disabled before..
    data['wepEnabled0'] = 'ON'
    data['length0'] = 1 if settings.security_type == settings.__class__.SECURITY_TYPE_WEP64 else 2
    data['format0'] = 2 if settings.is_wep_password_in_hex else 1
    data['defaultTxKeyId0'] = 1 # which WEP key is the default
    data['key10'] = settings.password
    data['key20'] = ''
    data['key30'] = ''
    data['key40'] = ''

    return data




from info import RouterInfo
from dmz import DMZSettings
from stats import TrafficStats
from wireless import WirelessSettings
from dhcp import DHCPReservationList, DHCPReservationListItem, \
     DHCPServerSettings
from clients import ConnectedClientsList, ConnectedClientsListItem
